//+++2004-11-01
//    Copyright (C) 2004  Mike Rieker, Beverly, MA USA
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//---2004-11-01

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

// This class holds the memory for the simulator

public class BigEndianMemory extends JInternalFrame {

  ////////////////////
  // STATIC SECTION //
  ////////////////////

  private static int memorysize;     // number of bytes of memory = memorydata.length
  private static byte[] memorydata;  // the actual memory data array

  private static ArrayList iodevs;

  // Called at initialization time to set the size of memory

  public static void setsize (int size) {
    memorysize = size;
    memorydata = new byte[size];
    iodevs = new ArrayList ();
  }

  // Called by anyone who wants to know how much memory there is

  public static int getsize () {
    return (memorysize);
  }

  // Called by IODev constructor to register an IO device

  public static void addiodev (IODev iodev) {

    // Make sure it is sane.  The check for upper 4K is just to be nice
    // to assembler coders so they can use %g0 as a base register.  But
    // if removed, then a check should be made to make sure it doesn't
    // overlap the memory array.

    if ((iodev.base >= 0) || (iodev.base < 0xFFFFF000)) {
      throw new IllegalArgumentException ("IO register base " + HexTextField.hexstring (iodev.base) + " not in upper 4K");
    }

    // The register mask should be a power of two

    for (int i = -2048; iodev.mask != i; i /= 2) {
      if (i == -1) throw new IllegalArgumentException ("IO register mask " + iodev.mask
                                                       + " not negative power of 2");
    }

    // The base should be on a natural mask boundary

    if ((iodev.base & ~ iodev.mask) != 0) {
      throw new IllegalArgumentException ("IO register base " + HexTextField.hexstring (iodev.base)
                                          + " not on mask " + iodev.mask + " boundary");
    }

    // Make sure it doesn't overlap another IO device

    for (int i = iodevs.size (); -- i >= 0;) {
      IODev iod = (IODev) iodevs.get (i);
      if (((iod.base ^ iodev.base) & iod.mask & iodev.mask) == 0) {
        throw new IllegalArgumentException ("IO register base " + HexTextField.hexstring (iodev.base)
                                            + " mask " + iodev.mask + " overlaps base "
                                            + HexTextField.hexstring (iod.base) + " mask " + iod.mask);
      }
    }

    // Ok, add it to the list

    iodevs.add (iodev);
  }

  // These routines read and write the memory and IO device arrays
  // Any writes also get posted to any memory display windows

  public static void write8 (int address, int data) {
    if (address >= 0) {
      memorydata[address] = (byte)data;
      SparcSim.updatememwindows (address & -4);
    } else {
      for (int i = iodevs.size (); -- i >= 0;) {
        IODev iodev = (IODev) iodevs.get (i);
        if (((iodev.base ^ address) & iodev.mask) == 0) {
          iodev.writereg (address - iodev.base, 1, data);
          break;
        }
      }
    }
  }

  public static void write16 (int address, int data) {
    if ((address & 1) != 0) throw new IllegalArgumentException ("Unaligned address");
    if (address >= 0) {
      memorydata[address+0] = (byte)(data >> 8);
      memorydata[address+1] = (byte)data;
      SparcSim.updatememwindows (address & -4);
    } else {
      for (int i = iodevs.size (); -- i >= 0;) {
        IODev iodev = (IODev) iodevs.get (i);
        if (((iodev.base ^ address) & iodev.mask) == 0) {
          iodev.writereg (address - iodev.base, 2, data);
          break;
        }
      }
    }
  }

  public static void write32 (int address, int data) {
    if ((address & 3) != 0) throw new IllegalArgumentException ("Unaligned address");
    if (address >= 0) {
      memorydata[address+0] = (byte)(data >> 24);
      memorydata[address+1] = (byte)(data >> 16);
      memorydata[address+2] = (byte)(data >> 8);
      memorydata[address+3] = (byte)data;
      SparcSim.updatememwindows (address);
    } else {
      for (int i = iodevs.size (); -- i >= 0;) {
        IODev iodev = (IODev) iodevs.get (i);
        if (((iodev.base ^ address) & iodev.mask) == 0) {
          iodev.writereg (address - iodev.base, 4, data);
          break;
        }
      }
    }
  }

  public static int read8 (int address) {
    if (address >= 0) {
      return (memorydata[address]);
    } else {
      for (int i = iodevs.size (); -- i >= 0;) {
        IODev iodev = (IODev) iodevs.get (i);
        if (((iodev.base ^ address) & iodev.mask) == 0) {
          return (iodev.readreg (address - iodev.base, 1));
        }
      }
      return (0);
    }
  }

  public static int read16 (int address) {
    if (address >= 0) {
      return ((((int)(memorydata[address+0])) << 8)
            + (((int)(memorydata[address+1])) & 0xFF));
    } else {
      for (int i = iodevs.size (); -- i >= 0;) {
        IODev iodev = (IODev) iodevs.get (i);
        if (((iodev.base ^ address) & iodev.mask) == 0) {
          return (iodev.readreg (address - iodev.base, 2));
        }
      }
      return (0);
    }
  }

  public static int read32 (int address) {
    if (address >= 0) {
      return ((((int)(memorydata[address+0])) << 24)
           + ((((int)(memorydata[address+1])) & 0xFF) << 16)
           + ((((int)(memorydata[address+2])) & 0xFF) << 8)
           + ((((int)(memorydata[address+3])) & 0xFF)));
    } else {
      for (int i = iodevs.size (); -- i >= 0;) {
        IODev iodev = (IODev) iodevs.get (i);
        if (((iodev.base ^ address) & iodev.mask) == 0) {
          return (iodev.readreg (address - iodev.base, 4));
        }
      }
      return (0);
    }
  }

  ////////////////////
  // OBJECT SECTION //
  ////////////////////

  private BigEndianMemory memoryframe;  // points to 'this'
  private HexTextField startfield;      // hexadecimal text field for start address
  private HexTextField endfield;        // hexadecimal text field for end address
  private JPanel displaypanel;          // all the data boxes (hexdatafields,chrdatafields)

  // Access to hexdatafields,chrdatafields,startaddr,endaddr
  //   must be synchronized by the 'memoryframe' object

  private HexTextField[] hexdatafields;  // 8-char hexadecimal data text fields
  private ChrMemData[] chrdatafields;    // 8-char character data text fields
  private int startaddr;                 // start address being displayed (16-byte aligned)
  private int endaddr;                   // end address (inclusive, 16-byte aligned)

  // Constructor - format the frame and make it visible

  public BigEndianMemory () {
    super ("Memorie", true, true, true, true);
    memoryframe = this;
    JPanel parameterpanel = new JPanel ();
    parameterpanel.setLayout (new GridLayout (2, 2));

    parameterpanel.add (new JLabel ("Start:"));
    startfield = new HexTextField (8);
    startfield.addFocusListener (new StartEndFocusListener ());
    parameterpanel.add (startfield);

    parameterpanel.add (new JLabel ("End:"));
    endfield = new HexTextField (8);
    endfield.addFocusListener (new StartEndFocusListener ());
    parameterpanel.add (endfield);

    displaypanel = new JPanel ();

    JPanel memorypanel = new JPanel ();
    memorypanel.setLayout (new FlowLayout ());
    memorypanel.add (parameterpanel);
    memorypanel.add (displaypanel);

    this.getContentPane ().add (memorypanel);
    this.pack ();
    this.setVisible (true);
  }

  // Called by MainFrame.updatememwindows to update this window's
  // display of the data at the given 32-bit aligned address

  public void updatedisplay (int address) {
    synchronized (memoryframe) {                       // make sure startaddr and array sizes don't change
      if (hexdatafields != null) {
        int i = (address - startaddr) / 4;
        if ((i >= 0) && (i < hexdatafields.length)) {  // see if address is within array limits
          int data = read32 (address);                 // if so, update data boxes on screen
          hexdatafields[i].setBinary (data);
          chrdatafields[i/2].setData (address & -8);
        }
      }
    }
  }

  // This is called when there is a change in a Start or End address box
  // We set up the rest of the window to display the contents of the memory

  private class StartEndFocusListener extends FocusAdapter {
    public void focusLost (FocusEvent fe) {

      // Get values entered in the two boxes

      int newstartaddr = startfield.getBinary () & -16;
      int newendaddr   = endfield.getBinary   () |  15;

      // Clip them to memory size limits
      // Round them to 16-byte boundaries

      if (newstartaddr < 0) newstartaddr = 0;
      if (newstartaddr >= memorysize) newstartaddr = memorysize - 16;
      if (newendaddr < newstartaddr) newendaddr = newstartaddr | 15;
      if (newendaddr >= memorysize) newendaddr = memorysize - 1;

      // Also don't allow over 1/4K per window

      if ((newendaddr - newstartaddr) >= 256) newendaddr = newstartaddr + 255;

      // See if there is a change before doing anything

      if ((newstartaddr != startaddr) || (newendaddr != endaddr)) {

        // There is a change, make sure nothing is playing with the arrays

        synchronized (memoryframe) {

          // Save new rounded values

          startaddr = newstartaddr;
          endaddr = newendaddr;
          startfield.setBinary (startaddr);
          endfield.setBinary (endaddr);

          // Clear out whatever used to be displayed

          displaypanel.removeAll ();

          // Set up new arrays for the new range

          int nbytes = endaddr + 1 - startaddr;
          GridLayout dagl = new GridLayout (nbytes / 16 + 1, 7);
          displaypanel.setLayout (dagl);
          hexdatafields = new HexTextField[nbytes/4];
          chrdatafields = new ChrMemData[nbytes/8];

          // Format an heading line to give addresses of the individual bytes in each line

          displaypanel.add (new FFJLabel ("        "));
          displaypanel.add (new FFJLabel ("0 1 2 3 "));
          displaypanel.add (new FFJLabel ("4 5 6 7 "));
          displaypanel.add (new FFJLabel ("8 9 A B "));
          displaypanel.add (new FFJLabel ("C D E F "));
          displaypanel.add (new FFJLabel ("01234567"));
          displaypanel.add (new FFJLabel ("89ABCDEF"));

          // Format the data text boxes and fill with current memory contents

          int j = 0; // hexdatafields array index
          int k = 0; // chrdatafields array index
          for (int address = startaddr; address <= endaddr; address += 16) {
            JLabel addresslabel = new JLabel (HexTextField.hexstring (address) + ":");
            addresslabel.setHorizontalAlignment (JLabel.RIGHT);
            displaypanel.add (addresslabel);
            for (int i = 0; i < 16; i += 4) {
              hexdatafields[j] = new HexTextField(8);
              hexdatafields[j].addFocusListener (new DataFieldFocusListener (j));
              hexdatafields[j].setBinary (read32 (address + i));
              displaypanel.add (hexdatafields[j++]);
            }
            for (int i = 0; i < 16; i += 8) {
              chrdatafields[k] = new ChrMemData (address + i);
              displaypanel.add (chrdatafields[k++]);
            }
          }

          // Re-size the window to accomodate the new memory display

          memoryframe.pack ();
        }
      }
    }
  }

  // This is called when there is a change in a data box
  // We simply write the value to memory
  // Use the write32 method so any other memory windows get updated

  private class DataFieldFocusListener extends FocusAdapter {
    int index;

    // Constructor - i is the index in the datafields array
    //               for the HexTextField being monitored

    public DataFieldFocusListener (int i) {
      index = i;
    }

    // This routine is called when exiting the text box,
    // presumably after something therein was modified

    // Write the new data value to memory via write32
    // so all views of the data get updated

    public void focusLost (FocusEvent fe) {
      int address = startaddr + index * 4;
      int olddata = read32 (address);
      int newdata = hexdatafields[index].getBinary ();
      if (newdata != olddata) write32 (address, newdata);
    }
  }

  // This class is just like a JLabel except it only does 8-char strings
  // and it will convert any non-standard ASCII characters to a '.'

  private class ChrMemData extends FFJLabel {
    public ChrMemData (int addr) {
      super ("        ");
      this.setData (addr);
    }
    public void setData (int addr) {
      char[] chrdata = new char[8];
      for (int i = 0; i < 8; i ++) {
        char c = (char) memorydata[addr+i];
        if ((c < ' ') || (c >= 127)) c = '.';
        chrdata[i] = c;
      }
      super.setText (new String (chrdata));
    }
  }

  // Just like a JLabel, except it is fixed font

  private class FFJLabel extends JLabel {
    public FFJLabel (String value) {
      super (value);
      super.setFont (SparcSim.monofont);
      super.setHorizontalAlignment (JLabel.RIGHT);
    }
  }
}
