//+++2004-11-01
//    Copyright (C) 2004  Mike Rieker, Beverly, MA USA
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//---2004-11-01

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

// This is just like an JTextField except it only allows hexadecimal characters
// It also provides getBinary and setBinary methods to fetch and store the value

public class HexTextField extends JTextField implements FocusListener {
  private int binaryvalue;
  private int ndigits;
  private String oldstring;

  // Constructor - must be supplied with number of digits

  public HexTextField (int nd) {
    super (nd + 1); // +1 because box comes out too small on some systems
    ndigits = nd;
    super.setFont (SparcSim.monofont);
    super.setHorizontalAlignment (JTextField.TRAILING);
    super.setText (hexstring (0, nd));
    super.addKeyListener (new HexKeyListener ());
    super.addFocusListener (this);
  }

  // Retrieve the binary value represented by hex digits currently in box

  public int getBinary () {
    this.focusLost (null);
    return (binaryvalue);
  }

  // Set binary value and update hexadecimal digits in box

  public void setBinary (int bv) {
    binaryvalue = bv;
    this.setString (hexstring (bv, ndigits));
  }

  private void setString (String str) {
    oldstring = str;
    super.setText (str);
  }

  // Don't let a caller set an arbitrary text string

  public void setText (String s) {
    throw new IllegalArgumentException ("cannot set to arbitrary string, use setBinary");
  }

  // We don't care about someone clicking in to the text box

  public void focusGained (FocusEvent fe) { }

  // This routine is called when the cursor is clicked out of the text box
  // We presume the text box was just modified with a new hexadecimal value
  // ... and so we update 'binaryvalue' accordingly

  public void focusLost (FocusEvent fe) {
    char c;
    int i, value;
    String strfield;

    // Try to convert string field from ascii hex to binary

    strfield = getText ();
    value = 0;
    for (i = 0; i < strfield.length (); i ++) {
      c = strfield.charAt (i);
      if ((c >= '0') && (c <= '9')) value = value * 16 + c - '0';
      else if ((c >= 'A') && (c <= 'F')) value = value * 16 + c - 'A' + 10;
      else if ((c >= 'a') && (c <= 'f')) value = value * 16 + c - 'a' + 10;
      else {
        JOptionPane.showMessageDialog (null, "Invalid hexadecimal number");
        i = -1;
        break;
      }
    }

    // If conversion is successful, update our 'permanent' binary value

    if (i > 0) binaryvalue = value;

    // If binaryvalue is different than what's on screen, update screen

    String canonstring = hexstring (binaryvalue, ndigits);
    if (!strfield.equals (canonstring)) {
      setString (canonstring);
    }
  }

  // This class will allow only up to ndigits hexadecimal characters in a hex text box

  private class HexKeyListener implements KeyListener {
    public void keyTyped    (KeyEvent e) { keyhit (); }
    public void keyPressed  (KeyEvent e) { keyhit (); }
    public void keyReleased (KeyEvent e) { keyhit (); }
  }

  private void keyhit () {
    String str = getText ();                       // get updated text string
    int cp = getCaretPosition ();                  // get caret position
    int i;
    int l = str.length ();                         // get updated string length
    StringBuffer sb;

    while ((l > ndigits) && (str.charAt (0) == '0')) {
      str = str.substring (1, l);                  // discard leading zeroes if too long
      -- l;
      if (cp > 0) -- cp;
    }

    if (l <= ndigits) {                            // make sure it's not too int
      sb = new StringBuffer (str);                 // ok, scan it
      for (i = l; -- i >= 0;) {
        char c = str.charAt (i);
        if ((c >= '0') && (c <= '9')) continue;    // check for valid hex digits
        if ((c >= 'A') && (c <= 'F')) continue;
        if ((c <  'a') || (c >  'f')) break;
        sb.setCharAt (i, (char) (c - 'a' + 'A'));  // ... make them uppercase
      }
      if (i < 0) {                                 // see if successful
        str = sb.toString ();                      // ok, update screen if different
        if (!str.equals (oldstring)) {
          setString (str);
          setCaretPosition (cp);
        }
        return;
      }
    }
    setString (oldstring);                         // something bad, restore old string
    setCaretPosition (-- cp);                      // restore caret position
    Toolkit.getDefaultToolkit ().beep ();          // beep
  }

  // General purpose routine to convert a binary value to an
  // hexadecimal string with leading zeroes (and capital letters)

  public static String hexstring (int value) {
    return (hexstring (value, 8));
  }

  public static String hexstring (int value, int digits) {
    char[] buffer = new char[digits];
    int digit;

    for (int i = 0; i < digits; i ++) {
      digit = (value >> ((digits * 4) - (i * 4) - 4)) & 15;
      if (digit < 10) buffer[i] = (char)('0' + digit);
      else buffer[i] = (char)('A' + digit - 10);
    }

    return (new String (buffer));
  }
}
