//+++2004-11-01
//    Copyright (C) 2004  Mike Rieker, Beverly, MA USA
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//---2004-11-01

// Prints various forms of data to an internal frame titled 'IODevPrint'
//  The CPU does a write to the following addresses to print:
//   0xFFFE0000 <- address of null-terminated ascii string
//   0xFFFE0004 <- hexadecimal integer
//   0xFFFE0008 <- signed decimal integer
//   0xFFFE000C <- unsigned decimal integer
//   0xFFFE0010 <- address of printf arg list
//                 +0: address of format string
//                 +4: address of first arg
//                 +8: address of second arg
//                +12: ...

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;

public class IODevPrint extends IODev {
  private JInternalFrame printframe;
  private JTextArea printarea;
  private boolean visible;

  public IODevPrint () {
    super (-32, 0xFFFFF000);  // register space is 32 bytes starting at 0xFFFFF000

    SparcAsm.defpermasym ("IODevPrintStrz",  super.base +  0);
    SparcAsm.defpermasym ("IODevPrintHex",   super.base +  4);
    SparcAsm.defpermasym ("IODevPrintInt",   super.base +  8);
    SparcAsm.defpermasym ("IODevPrintUint",  super.base + 12);
    SparcAsm.defpermasym ("IODevPrintfArgs", super.base + 16);

    // Build the textarea to hold the text itself and put a scroll box around it

    printarea = new JTextArea (20, 60);
    printarea.setFont (SparcSim.monofont);
    printarea.setEditable (false);
    JScrollPane printscroll = new JScrollPane (printarea);

    // Set up a Save button so user can easily save text to a file

    JButton savebutton = new JButton ("Save As");
    savebutton.addActionListener (new SaveButtonListener ());

    // Set up a Clear button so user can easily clear the text box

    JButton clearbutton = new JButton ("Clear");
    clearbutton.addActionListener (new ClearButtonListener ());

    // Create internal frame for the text area and clear button
    // Don't put a close button (X) in upper right because it can't be re-opened

    printframe = new JInternalFrame ("IODevPrint", true, false, true, true);
    JPanel printpanel = (JPanel) printframe.getContentPane ();

    // Text area takes up all that's available except for only what buttons need

    GridBagLayout gbl = new GridBagLayout ();
    GridBagConstraints gbc = new GridBagConstraints ();

    printpanel.setLayout (gbl);

    gbc.weightx = 1.0;
    gbc.gridwidth = GridBagConstraints.REMAINDER;

    gbc.fill = GridBagConstraints.BOTH;
    gbc.gridheight = GridBagConstraints.RELATIVE;
    gbc.weighty = 1.0;
    gbl.setConstraints (printscroll, gbc);
    printpanel.add (printscroll);

    gbc.fill = GridBagConstraints.NONE;
    gbc.gridheight = GridBagConstraints.REMAINDER;

    JPanel buttonpanel = new JPanel ();
    buttonpanel.setLayout (new FlowLayout ());
    buttonpanel.add (savebutton);
    buttonpanel.add (clearbutton);

    gbc.weighty = 0.0;
    gbl.setConstraints (buttonpanel, gbc);
    printpanel.add (buttonpanel);

    //printframe.setSize (sourcetextwidth * (linenumberspaces.length () + 1 + assemblerlabelwidth + defaulttextareawidth) + sourcetextheight,
    //                     sourcetextheight * defaulttextarealines + 50);

    printframe.pack ();

    // Leave it invisible until the first thing is written to it

    visible = false;
    SparcSim.addtomainpane (printframe);
  }

  // There is nothing to read

  public int readreg (int reg, int size) {
    return (0);
  }

  public void writereg (int reg, int size, int data) {
    if (!visible) {
      printframe.setVisible (true);
      visible = true;
    }

    switch (reg) {
      case 0: {
        char c;
        StringBuffer s = new StringBuffer ();

        while ((c = (char) BigEndianMemory.read8 (data ++)) != 0) {
          s.append (c);
        }
        printarea.append (s.toString ());
        break;
      }
      case 4: {
        printarea.append (HexTextField.hexstring (data, size * 2));
        break;
      }
      case 8: {
        if (size == 1) data = (byte) data;
        if (size == 2) data = (short) data;
        printarea.append (Integer.toString (data));
        break;
      }
      case 12: {
        if (size == 1) data &= 0xFF;
        if (size == 2) data &= 0xFFFF;
        if (data >= 0) printarea.append (Integer.toString (data));
        else {
          int billions = 0;
          do {
            data -= 1000000000;
            billions ++;
          } while ((data < 0) || (data >= 1000000000));
          printarea.append (Integer.toString (billions) + Integer.toString (1000000000 + data).substring (1, 10));
        }
        break;
      }
      case 16: {
        doprintf (data);
        break;
      }
    }
  }

  // Internal routine to handle printf-like request
  //  Input:
  //   address = address in simulated memory of printf argument list
  //             @address = pointer to null-terminated format string

  private void doprintf (int address) {
    char fc;
    Par par = new Par ();

    par.fp = BigEndianMemory.read32 (address);				// set up format string pointer
    par.ap = address + 4;						// set up argument list pointer

    while ((fc = (char) BigEndianMemory.read8 (par.fp)) != 0) {		/* scan each character in the format string */
      par.fp ++;
      if (fc != '%') {							/* if not a percent, copy as is to output buffer */
        if (fc == '\n') par.putch ('\r');
        par.putch (fc);
        continue;
      }

      par.altform   = false;						/* percent, initialize param block values */
      par.zeropad   = false;
      par.leftjust  = false;
      par.posblank  = false;
      par.plussign  = false;
      par.minwidth  = 0;
      par.precision = -1;
      //par.intsize   = 0;

    /* Process format item prefix characters */

processitemprefixes:
      while ((fc = (char) BigEndianMemory.read8 (par.fp)) != 0) {
        par.fp ++;
        switch (fc) {
          case '#': par.altform = true; break;				/* 'alternate form' */
          case '0': if (!par.leftjust) par.zeropad  = true; break;	/* zero pad the numeric value (ignore precision) */
          case '-': par.leftjust = true;  par.zeropad  = false; break;	/* left justify space fill (otherwise it is right justified) */
          case ' ': if (!par.plussign) par.posblank = true; break;	/* leave a space for a plus sign if value is positive */
          case '+': par.plussign = true;  par.posblank = false; break;	/* include a '+' if value is positive */
          case '1':
          case '2':
          case '3':
          case '4':
          case '5':
          case '6':
          case '7':
          case '8':
          case '9':
          case '*': -- par.fp; par.minwidth = par.getfmtint (); break;	/* minimum field width */
          case '.': par.precision = par.getfmtint (); break;		/* precision */

          default: break processitemprefixes;				/* unknown, assume it's the final character */
        }
      }

      /* Process final format character using those prefixes */

      switch (fc) {

        /* Character and string convertions */

        case 'c': {					/* - single character */
          fc = (char) BigEndianMemory.read8 (par.ap);		        /* get character to be output */
          par.ap += 4;
          if (!(par.leftjust)) par.putfc (par.minwidth - 1, ' ');	/* output leading padding */
          par.putch (fc);						/* output the character */
          par.putrightfill ();						/* output trailing padding */
          break;
        }

        case 's': {					/* - character string */
          int i, sp;
          StringBuffer sb = new StringBuffer ();

          sp = BigEndianMemory.read32 (par.ap);			/* point to string to be output */
          par.ap += 4;
          if (par.precision < 0) {
            if (sp == 0) i = 5;
            else for (i = 0; BigEndianMemory.read8 (sp + i) != 0; i ++) {}
          } else {
            for (i = 0; i < par.precision; i ++) {			/* ... careful not to step too far */
              if (BigEndianMemory.read8 (sp + i) == 0) break;
            }
          }
          if ((par.precision < 0) || (par.precision > i)) par.precision = i; /* get length to be output */
          if (!(par.leftjust)) par.putfc (par.minwidth - par.precision, ' '); /* output leading padding */
          if (sp == 0) sb.append ("(nil)");				/* get the string */
          else {
            for (i = 0; i < par.precision; i ++) {
              sb.append ((char) BigEndianMemory.read8 (sp + i));
            }
          }
          par.putst (par.precision, sb.toString ());			/* output the string itself */
          par.putrightfill ();						/* output trailing padding */
          break;
        }

        /* Integer convertions */

        case 'D': //par.intsize = 4;
        case 'd':
        case 'i': {					/* - signed decimal */
          par.getsignednum ();						/* get value into 'signednum', with sign in 'negative' */
          par.nci = par.ncb.length;					/* point to end of numeric conversion buffer */
          do {								/* convert to decimal string in end of ncb */
            par.ncb[--par.nci] = (char) ((par.unsignednum % 10) + '0');
            par.unsignednum /= 10;
          } while (par.unsignednum != 0);
          par.putint ("");
          break;
        }

        case 'O': //par.intsize = 4;
        case 'o': {					/* - unsigned octal */
          par.getunsignednum ();					/* get value into 'unsignednum' */
          par.nci = par.ncb.length;					/* point to end of numeric conversion buffer */
          do {								/* convert to octal string in end of ncb */
            par.ncb[--par.nci] = (char) ((par.unsignednum & 7) + '0');
            par.unsignednum = (par.unsignednum >> 3) & 0x1FFFFFFF;
          } while (par.unsignednum != 0);
          par.putint ("0");
          break;
        }

        case 'U': //par.intsize = 4;
        case 'u': {					/* - unsigned decimal */
          int billions = 0;

          par.getunsignednum ();					/* get value into 'unsignednum' */
          if (par.unsignednum < 0) {                                    /* idiot java */
            do {
              billions ++;
              par.unsignednum -= 1000000000;
            } while ((par.unsignednum < 0) || (par.unsignednum > 999999999));
            par.unsignednum += 1000000000;
          }
          par.nci = par.ncb.length;					/* point to end of numeric conversion buffer */
          do {								/* convert to decimal string in end of ncb */
            par.ncb[--par.nci] = (char) ((par.unsignednum % 10) + '0');
            par.unsignednum /= 10;
          } while (par.unsignednum != 0);
          if (billions > 0) par.ncb[par.nci] = (char) (billions + '0');
          par.putint ("");
          break;
        }

        case 'x': {					/* - unsigned hexadecimal (lower case) */
          par.getunsignednum ();					/* get value into 'unsignednum' */
          par.nci = par.ncb.length;					/* point to end of numeric conversion buffer */
          do {								/* convert to decimal string in end of ncb */
            par.ncb[--par.nci] = "0123456789abcdef".charAt (par.unsignednum & 15);
            par.unsignednum = (par.unsignednum >> 4) & 0x0FFFFFFF;
          } while (par.unsignednum != 0);
          par.putint ("0x");
          break;
        }

        case 'X': {					/* - unsigned hexadecimal (caps) */
          par.getunsignednum ();					/* get value into 'unsignednum' */
          par.nci = par.ncb.length;					/* point to end of numeric conversion buffer */
          do {								/* convert to decimal string in end of ncb */
            par.ncb[--par.nci] = "0123456789ABCDEF".charAt (par.unsignednum & 15);
            par.unsignednum = (par.unsignednum >> 4) & 0x0FFFFFFF;
          } while (par.unsignednum != 0);
          par.putint ("0x");
          break;
        }

        /* Misc convertions */

        case 'p': {					/* - pointer */
          par.getunsignednum ();					/* get value */
          par.negative = false;						/* it's never negative by definition */
          par.nci = par.ncb.length;					/* point to end of numeric conversion buffer */
          do {								/* convert to decimal string in end of ncb */
            par.ncb[--par.nci] = "0123456789abcdef".charAt (par.unsignednum & 15);
            par.unsignednum = (par.unsignednum >> 4) & 0x0FFFFFFF;
          } while (par.unsignednum != 0);
          par.altform = true;						/* always include the 0x */
          par.putint ("0x");
          break;
        }

        case 'n': {					/* - number of output characters so far */
          int outaddr;

          outaddr = BigEndianMemory.read32 (par.ap);
          par.ap += 4;
          BigEndianMemory.write32 (outaddr, par.numout);
          break;
        }

        /* Don't know what it is, just output it as is (maybe it was '%%') */

        default: {
          par.putch (fc);
          break;
        }
      }
    }

    if (par.oi > 0) printarea.append (new String (par.ob, 0, par.oi));
  }

  /* Internal parameter block */

  private class Par {

    int fp;		// format string pointer

    int ap;		// argument list pointer
    int signednum;	// last fetched signed number
    int unsignednum;	// last fetched unsigned number

    int oi;		// index in 'ob' for next output character
    char[] ob;		// array of output characters

    int minwidth;		// minimum output field width
    int precision;	// number of decimal places to output
    boolean negative;	// number being output requires a '-' sign
    boolean altform;	// output it in alternative format
    boolean zeropad;	// right justify, zero fill
    boolean leftjust;	// left justify, blank fill
    boolean posblank;	// if positive, output a space for the sign
    boolean plussign;	// if positive, output a '+' for the sign

    int nci;
    char[] ncb;		// integer string to output

    int numout;		// total number of characters output so far

    // Constructor allocates output char buffer

    public Par () {
      oi = 0;
      ob = new char[256];
      ncb = new char[32];
    }

    // Get a signed number from the argument list

    public void getsignednum () {
      this.signednum = BigEndianMemory.read32 (this.ap);
      this.ap += 4;
      this.negative = (this.signednum < 0);
      this.unsignednum = this.signednum;
      if (this.signednum < 0) this.unsignednum = - this.signednum;
    }

    // Get an unsigned number from the argument list

    public void getunsignednum () {
      this.unsignednum = BigEndianMemory.read32 (this.ap);
      this.ap += 4;
      this.negative = false;
    }

    /* Get integer in format string.  If *, use next integer argument, otherwise use decimal string */

    public int getfmtint () {
      char fc;
      int val;

      if (BigEndianMemory.read8 (this.fp) == '*') {
        this.fp ++;
        val = BigEndianMemory.read32 (this.ap);
        this.ap += 4;
      } else {
        val = 0;
        while (((fc = (char) BigEndianMemory.read8 (this.fp)) >= '0') && (fc <= '9')) {
          val = val * 10 + fc - '0';
          this.fp ++;
        }
      }
      return (val);
    }

    /* Put the integer in par.nc, optionally prepending the alternate string and padding as required */

    public void putint (String alt)

    {
      int altneeded, numdigs, numzeroes, signneeded;

      numdigs = this.ncb.length - this.nci;				/* get number of digits in par.ncb string */

      signneeded = 0;
      if (this.negative || this.plussign || this.posblank) signneeded = 1; /* maybe another char will be required for sign position */

      altneeded = 0;							/* maybe need the alternate string */
      if (this.altform) altneeded = alt.length ();

      numzeroes = 0;  							/* maybe it needs zeroes on front for precision padding */
      if (numdigs < this.precision) numzeroes = this.precision - numdigs;

      putleftfill (signneeded + altneeded + numzeroes + numdigs);	/* output any left blank filling */
      putsign ();							/* output the sign character, if any */
      putst (altneeded, alt);						/* output the alternate string, if any */
      putleftzero (numzeroes + numdigs);				/* put left zero padding (for minimum width) */
      putfc (numzeroes, '0');						/* put left zeroes for precision */
      putst (numdigs, new String (this.ncb, this.nci, numdigs));	/* output the digits as a string */
      putrightfill ();							/* pad on the right with blanks if needed (for minimum width) */
    }

    /* Put Left Fill */

    public void putleftfill (int n) {
      if (!this.leftjust && !this.zeropad) putfc (this.minwidth - n, ' ');
    }

    public void putleftzero (int n) {
      if (!this.leftjust &&  this.zeropad) putfc (this.minwidth - n, '0');
    }

    public void putrightfill () {
      putfc (this.minwidth, ' ');
    }

    /* Output sign character, given par.negative, plussign, posblank */

    public void putsign ()

    {
      if (this.negative) putch ('-');
      else if (this.plussign) putch ('+');
      else if (this.posblank) putch (' ');
    }

    /* Output fill characters, flush if full */

    public void putfc (int s, char c)

    {
      if (s > 0) {
        this.numout   += s;						/* this much more will have been output */
        this.minwidth -= s;						/* this fewer chars needed to fill output field minimum width */

        while (-- s > 0) {
          if (this.oi == this.ob.length) {				/* see if any room left */
            printarea.append (new String (this.ob));			/* if not, output buffer as is */
            this.oi = 0;						/* ... then reset index */
          }
          this.ob[this.oi++] = c;					/* put fill char in buffer */
        }
      }
    }

    /* Output string, flush if full */

    public void putst (int s, String b)

    {
      if (s > 0) {
        this.numout   += s;						/* this much more will have been output */
        this.minwidth -= s;						/* this fewer chars needed to fill output field minimum width */
        for (int i = 0; i < s; i ++) {
          if (this.oi == this.ob.length) {				/* see if buffer is full */
            printarea.append (new String (this.ob));			/* output whole buffer */
            this.oi = 0;						/* reset output index */
          }
          this.ob[this.oi++] = b.charAt (i);				/* fill it with parameter */
        }
      }
    }

    /* Output single character, flush if full */

    public void putch (char c)

    {
      int siz;

      if (this.oi == this.ob.length) {					/* see if buffer is full */
        printarea.append (new String (this.ob));			/* output whole buffer */
        this.oi = 0;							/* reset output index */
      }
      this.ob[this.oi++] = c;						/* store character in buffer */
      this.numout ++;							/* one more char has been output */
      this.minwidth --;							/* one less char needed to fill output field minimum width */
    }
  }

  // Save button clicked - save text area to a file

  private File savefile = null;

  private class SaveButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {

      // Ask where they want the file saved

      JFileChooser filechooser = new JFileChooser ();
      if (savefile != null) filechooser.setSelectedFile (savefile);
      filechooser.setDialogTitle ("Save IODevPrint text to file");
      int rc = filechooser.showSaveDialog (printframe);
      if (rc == JFileChooser.APPROVE_OPTION) {
        FileOutputStream fileoutputstream = null;

        // Write text from the textarea to the file
        // But if there's an existing file with the same name, rename it first

        savefile = filechooser.getSelectedFile();
        try {
          int version = 0;

          while (!savefile.createNewFile ()) {
            File newname = new File (savefile.getAbsolutePath () + "." + (++ version) + "~");
            if (!newname.exists () && savefile.renameTo (newname)) break;
          }
          fileoutputstream = new FileOutputStream (savefile);
        }
        catch (IOException ioe) {
          JOptionPane.showMessageDialog(null,
                                        "Error creating " + savefile.getAbsolutePath() +
                                        ": " + ioe.getMessage());
          fileoutputstream = null;
        }
        if (fileoutputstream != null) {
          OutputStreamWriter outputstreamwriter = new OutputStreamWriter (fileoutputstream);
          try {
            outputstreamwriter.write (printarea.getText ());
            outputstreamwriter.close ();
          }
          catch (IOException ioe) {
            JOptionPane.showMessageDialog(null,
                                          "Error writing " +
                                          savefile.getAbsolutePath() +
                                          ": " + ioe.getMessage());
          }
        }
      }
    }
  }

  // Clear button clicked - clear the text area

  private class ClearButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      printarea.setText ("");
    }
  }
}
