//+++2004-11-01
//    Copyright (C) 2004  Mike Rieker, Beverly, MA USA
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//---2004-11-01

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;

// These objects are the Source File Editor frames
// They will edit and compile a source file
// They also have breakpoint buttons on them for source lines

public class SourceEditor extends JInternalFrame {

  private JInternalFrame sourceframe;
  private Container sourcepanel;
  private BptButton[] bptbuttonarray;
  private JLabel[] assemblerlabelarray;
  private JTextArea sourcetextarea;
  private JPanel bptbuttonpanel;
  private JPanel assemblerlabelpanel;
  private JPanel linenumberpanel;
  private int linecount;
  private int loaddress, hiaddress;

  private int sourcetextheight, sourcetextwidth;  // height and width of text characters (monospaced font)

  private final String linenumberspaces = "    "; // number of spaces for the source file line numbers
  private final int assemblerlabelwidth = 20;     // number of spaces for the assembler object code / error messages
  private final int defaulttextareawidth = 60;    // default number of columns for source code window
  private final int defaulttextarealines = 20;    // default number of lines for source code window

  private JButton openbutton;                     // the various buttons
  private JButton savebutton;
  private JButton compilebutton;

  private boolean changesmade;                    // set 'true' when a change is made
                                                  // set 'false' when file just opened or saved
  private File sourcefile;                        // name of file when last opened or saved

  private static File lastmanualsavedirectory;    // which directory was last selected in any manual save operation
  private static File lastmanualopendirectory;    // which directory was last selected in any manual open operation

  public SourceEditor () {
    super ("Source Editor", true, true, true, true);

    // Create a window to put stuff on

    sourceframe = this;
    sourcepanel = sourceframe.getContentPane ();

    // Create breakpoint buttons, assembler output, line numbers, text window
    // None of these things contain scrollbars, that is applied externally to them as a whole
    // There must be one { breakpoint button, assembler label, line number } per line in the source file
    // The source text area must encompass the entire source file

    bptbuttonpanel = new JPanel ();
    assemblerlabelpanel = new JPanel ();
    linenumberpanel = new JPanel ();

    sourcetextarea = new JTextArea ();
    sourcetextarea.setFont (SparcSim.monofont);
    sourcetextarea.addKeyListener (new SourceTextListener ());

    // Measure the height of characters in the sourcetextarea
    // This is how we know how big to make the bptbuttons, etc, so they will match up

    sourcetextheight = sourceframe.getFontMetrics (SparcSim.monofont).getHeight ();
    sourcetextwidth  = sourceframe.getFontMetrics (SparcSim.monofont).charWidth ('m');

    // Make the scrollable panel that consists of linenumbers,
    // assembler labels, breakpoint buttons and source text area

    JScrollPane scrolledpanel;
    {
      JPanel fiddlybitspanel = new JPanel ();
      fiddlybitspanel.setLayout (new BoxLayout (fiddlybitspanel, BoxLayout.X_AXIS));
      fiddlybitspanel.add (linenumberpanel);
      fiddlybitspanel.add (assemblerlabelpanel);
      fiddlybitspanel.add (bptbuttonpanel);

      JPanel scrollablepanel = new JPanel ();
      GridBagLayout gbl = new GridBagLayout ();
      GridBagConstraints gbc = new GridBagConstraints ();
      scrollablepanel.setLayout (gbl);

      gbc.weighty = 1.0;
      gbc.gridheight = GridBagConstraints.REMAINDER;
      gbc.anchor = GridBagConstraints.NORTH;

      gbc.fill = GridBagConstraints.NONE;
      gbc.gridwidth = GridBagConstraints.RELATIVE;
      gbc.weightx = 0.0;
      gbl.setConstraints (fiddlybitspanel, gbc);
      scrollablepanel.add (fiddlybitspanel);

      gbc.fill = GridBagConstraints.BOTH;
      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gbc.weightx = 1.0;
      gbl.setConstraints (sourcetextarea, gbc);
      scrollablepanel.add (sourcetextarea);

      scrolledpanel = new JScrollPane (scrollablepanel);
    }

    // Create the function buttons

    openbutton = new JButton ("Open");
    openbutton.addActionListener (new OpenButtonListener ());
    savebutton = new JButton ("Save (As)");
    savebutton.addActionListener (new SaveButtonListener ());
    compilebutton = new JButton ("Assemble");
    compilebutton.addActionListener (new CompileButtonListener ());

    // Put the main window together
    // The scrolledpanel goes on top with the function buttons on the bottom

    {
      // Make a little panel with the buttons on it
      // They should end up horizontal

      JPanel buttonpanel = new JPanel ();
      buttonpanel.setLayout (new FlowLayout ());
      buttonpanel.add (openbutton);
      buttonpanel.add (savebutton);
      buttonpanel.add (compilebutton);

      // Use a gridbag to put a big area on top and little area on bottom

      GridBagLayout gbl = new GridBagLayout ();
      GridBagConstraints gbc = new GridBagConstraints ();
      sourcepanel.setLayout (gbl);

      gbc.weightx = 1.0;
      gbc.gridwidth = GridBagConstraints.REMAINDER;

      // Give the scrolledpanel most of the height (using weighty)

      gbc.fill = GridBagConstraints.BOTH;
      gbc.gridheight = GridBagConstraints.RELATIVE;
      gbc.weighty = 1.0;
      gbl.setConstraints (scrolledpanel, gbc);
      sourcepanel.add (scrolledpanel);

      // The buttons go on the bottom
      // Setting weighty = 0.0 gives them only what they need

      gbc.fill = GridBagConstraints.NONE;
      gbc.gridheight = GridBagConstraints.REMAINDER;
      gbc.weighty = 0.0;
      gbl.setConstraints (buttonpanel, gbc);
      sourcepanel.add (buttonpanel);
    }

    // Now the panel is all put together, make it visible

    sourceframe.setSize (sourcetextwidth * (linenumberspaces.length () + 1 + assemblerlabelwidth + defaulttextareawidth) + sourcetextheight,
                         sourcetextheight * defaulttextarealines + 50);
    sourceframe.addInternalFrameListener (new SourceEditorListener ());
    sourceframe.setVisible (true);

    changesmade = false;
    savebutton.setEnabled (false);
    sourcefile = null;
  }

  // This can be called to force a particular source file to be opened in the window

  public boolean openfile (String filename) {
    FileInputStream fileinputstream = null;

    if (!maybesavechanges (JOptionPane.YES_NO_CANCEL_OPTION)) return (false);

    // Open the source file

    sourcefile = new File (filename);
    try {
      fileinputstream = new FileInputStream (sourcefile);
    } catch (IOException ioe) {
      JOptionPane.showMessageDialog (null, "Error opening " + sourcefile.getAbsolutePath () + ": " + ioe.getMessage ());
      return (false);
    }

    // Read text from the file to the sourcetextarea

    InputStreamReader inputstreamreader = new InputStreamReader (fileinputstream);
    try {
      char[] buf = new char[4096];
      int len;

      sourcetextarea.setText ("");
      while ((len = inputstreamreader.read (buf, 0, buf.length)) > 0) {
        sourcetextarea.append (new String (buf, 0, len));
      }
      inputstreamreader.close ();
    }
    catch (IOException ioe) {
      JOptionPane.showMessageDialog (null,
                                     "Error reading " + sourcefile.getAbsolutePath () +
                                     ": " + ioe.getMessage ());
      return (false);
    }
    String absolutepath = sourcefile.getAbsolutePath ();
    sourceframe.setTitle (absolutepath);
    sourcefile = new File (absolutepath);
    changesmade = false;
    savebutton.setEnabled (false);
    return (true);
  }

  // If there are changes in the source that haven't been saved to a file
  //   ask the user if they want them saved.
  // Return false if the user says to 'cancel', which means don't do whatever
  //   is causing this window to close, but leave the window as it is with the
  //   current modified source code.
  // Return true if it is ok to go ahead and either close this window or replace
  //   the contents with something completely different.

  public boolean maybesavechanges (int yes_no_cancel) {
    if (!changesmade) return (true);

    String prompt = "Do you wish to save this new source file?";
    if (sourcefile != null) prompt = "Do you wish to save changes to " + sourcefile.getAbsolutePath () + "?";
    int rc = JOptionPane.showConfirmDialog (sourceframe, prompt, "Closing source editor", yes_no_cancel);

    switch (rc) {
      case JOptionPane.CANCEL_OPTION: {
        return (false);
      }
      case JOptionPane.NO_OPTION: {
        return (true);
      }
      case JOptionPane.YES_OPTION: {
        writesourcefile ();
        return (true);
      }
      default: throw new IllegalArgumentException ();
    }
  }

  // Write the source out to a file

  private void writesourcefile () {

    // Ask where they want the file saved (default is where it came from)

    JFileChooser filechooser = new JFileChooser ();
    if (sourcefile != null) filechooser.setSelectedFile (sourcefile);
    else if (lastmanualsavedirectory != null) filechooser.setSelectedFile (lastmanualsavedirectory);
    filechooser.setDialogTitle ("Save source to file");
    int rc = filechooser.showSaveDialog (sourceframe);
    if (rc == JFileChooser.APPROVE_OPTION) {
      FileOutputStream fileoutputstream = null;

      // Write text from the sourcetextarea to the file
      // But if there's an existing file with the same name, rename it first

      sourcefile = filechooser.getSelectedFile();
      lastmanualsavedirectory = sourcefile.getParentFile ();
      try {
        int version = 0;

        while (!sourcefile.createNewFile ()) {
          File newname = new File (sourcefile.getAbsolutePath () + "." + (++ version) + "~");
          if (!newname.exists () && sourcefile.renameTo (newname)) break;
        }
        fileoutputstream = new FileOutputStream (sourcefile);
      }
      catch (IOException ioe) {
        JOptionPane.showMessageDialog(null,
                                      "Error creating " + sourcefile.getAbsolutePath() +
                                      ": " + ioe.getMessage());
        fileoutputstream = null;
      }
      if (fileoutputstream != null) {
        OutputStreamWriter outputstreamwriter = new OutputStreamWriter(
            fileoutputstream);
        try {
          outputstreamwriter.write(sourcetextarea.getText());
          outputstreamwriter.close();
        }
        catch (IOException ioe) {
          JOptionPane.showMessageDialog(null,
                                        "Error writing " +
                                        sourcefile.getAbsolutePath() +
                                        ": " + ioe.getMessage());
        }
        sourceframe.setTitle (sourcefile.getAbsolutePath ());
        changesmade = false;
        savebutton.setEnabled (false);
      }
    }
  }

  private class SourceEditorListener extends InternalFrameAdapter {
    public void internalFrameClosing (InternalFrameEvent e) {
      if (changesmade) maybesavechanges (JOptionPane.YES_NO_OPTION);
    }
  }

  // Whenever changes are made to the source file,
  // wipe out our buttons, assembler output and linenumbers

  private class SourceTextListener extends KeyAdapter {
    public void keyTyped (KeyEvent e) {
      if (!changesmade) {
        bptbuttonpanel.removeAll ();
        assemblerlabelpanel.removeAll ();
        linenumberpanel.removeAll ();
        changesmade = true;
        savebutton.setEnabled (true);
      }
    }
  }

  private class OpenButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      JFileChooser filechooser = new JFileChooser ();
      if (sourcefile != null) filechooser.setSelectedFile (sourcefile);
      else if (lastmanualopendirectory != null) filechooser.setSelectedFile (lastmanualopendirectory);
      filechooser.setDialogTitle ("Open source file");
      int rc = filechooser.showOpenDialog (sourceframe);
      if (rc == JFileChooser.APPROVE_OPTION) {
        lastmanualopendirectory = filechooser.getSelectedFile ().getParentFile ();
        openfile (filechooser.getSelectedFile ().getAbsolutePath ());
      }
    }
  }

  // This is the Save button listener - called whenever the Save button is clicked

  private class SaveButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      writesourcefile ();
    }
  }

  // Retrieve name of source file being edited

  public String getsourcename () {
    if (sourcefile == null) return ("");
    return (sourcefile.getAbsolutePath ());
  }

  // Retrieve source text string

  public String gettextstring () {
    return (sourcetextarea.getText ());
  }

  // This is the Compile button listener
  // It recompiles everything because defined symbols may have moved

  private class CompileButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      SparcSim.compileall ();
    }
  }

  // Object code is about to be assembled -
  // Wipe out the breakpoint buttons, assembled object code
  // Re-initialize the arrays and re-display buttons

  public void resetasminfo () {

    linecount = sourcetextarea.getLineCount ();

    bptbuttonpanel.removeAll ();
    assemblerlabelpanel.removeAll ();
    linenumberpanel.removeAll ();

    if (linecount <= 0) return;

    bptbuttonpanel.setPreferredSize (new Dimension (sourcetextheight, linecount * sourcetextheight));
    linenumberpanel.setPreferredSize (new Dimension (sourcetextwidth * (linenumberspaces.length () + 1), linecount * sourcetextheight));
    assemblerlabelpanel.setPreferredSize (new Dimension (sourcetextwidth * assemblerlabelwidth, linecount * sourcetextheight));

    bptbuttonarray = new BptButton[linecount];
    assemblerlabelarray = new JLabel[linecount];

    SpringLayout bbsl = new SpringLayout ();
    bptbuttonpanel.setLayout (bbsl);

    SpringLayout alsl = new SpringLayout ();
    assemblerlabelpanel.setLayout (alsl);

    SpringLayout lnsl = new SpringLayout ();
    linenumberpanel.setLayout (lnsl);

    for (int i = 0; i < linecount; i ++) {
      bptbuttonarray[i] = new BptButton ();
      bptbuttonarray[i].setVisible (false);
      bbsl.putConstraint (SpringLayout.WEST, bptbuttonarray[i], 0, SpringLayout.WEST, bptbuttonpanel);
      bbsl.putConstraint (SpringLayout.NORTH, bptbuttonarray[i], i * sourcetextheight, SpringLayout.NORTH, bptbuttonpanel);
      bptbuttonpanel.add (bptbuttonarray[i]);

      assemblerlabelarray[i] = new JLabel ("");
      assemblerlabelarray[i].setFont (SparcSim.monofont);
      alsl.putConstraint (SpringLayout.WEST, assemblerlabelarray[i], 0, SpringLayout.WEST, assemblerlabelpanel);
      alsl.putConstraint (SpringLayout.NORTH, assemblerlabelarray[i], i * sourcetextheight, SpringLayout.NORTH, assemblerlabelpanel);
      assemblerlabelpanel.add (assemblerlabelarray[i]);

      String labelstring = Integer.toString (i + 1);
      JLabel linenumber = new JLabel (linenumberspaces.substring (labelstring.length ()) + labelstring);
      linenumber.setFont (SparcSim.monofont);
      lnsl.putConstraint (SpringLayout.WEST, linenumber, 0, SpringLayout.WEST, linenumberpanel);
      lnsl.putConstraint (SpringLayout.NORTH, linenumber, i * sourcetextheight, SpringLayout.NORTH, linenumberpanel);
      linenumberpanel.add (linenumber);
    }

    loaddress = hiaddress = 0;
  }

  // The assembler calls this routine back with all object code it compiles
  // We just display the address and data on the line if we haven't already put something there
  // Also, if this is the first time, we save the breakpoint address

  public void objcode (int linenumber, int address, int size, int data) {
    if (assemblerlabelarray[linenumber-1].getText ().equals ("")) {
      assemblerlabelarray[linenumber-1].setText (HexTextField.hexstring (address) + " "
                                                 + HexTextField.hexstring (data, size * 2));
    }
    if (bptbuttonarray[linenumber-1].address == 0) {
      if ((loaddress | hiaddress) == 0) {
        loaddress = hiaddress = address;
      } else if (address < loaddress) {
        loaddress = address;
      } else if (address > hiaddress) {
        hiaddress = address;
      }
      bptbuttonarray[linenumber-1].address = address;
      bptbuttonarray[linenumber-1].setVisible (true);
    }
  }

  // The assembler calls this routine back with an error message for a given line
  // If we haven't already displayed an error message for the line, display it in red

  public void errormsg (int linenumber, String message) {
    if (assemblerlabelarray[linenumber-1].getForeground () != Color.RED) {
      assemblerlabelarray[linenumber-1].setText (message);
      assemblerlabelarray[linenumber-1].setForeground (Color.RED);
      sourceframe.setVisible (true);
      sourceframe.moveToFront ();
    }
  }

  // This will turn the assembler labels green for only those addresses listed in addresses

  public void updatecpulocation (int[] addresses) {
    if (bptbuttonarray == null) return;
    for (int i = bptbuttonarray.length; -- i >= 0;) {
      Color newfgcolor = Color.BLACK;
      Color oldfgcolor = assemblerlabelarray[i].getForeground ();
      if (oldfgcolor == Color.RED) continue;
      for (int j = addresses.length; -- j >= 0;) {
        if (bptbuttonarray[i].address == addresses[j]) newfgcolor = Color.GREEN;
      }
      if (oldfgcolor != newfgcolor) {
        assemblerlabelarray[i].setForeground (newfgcolor);
      }
    }
  }

  // These are the breakpoint buttons

  private class BptButton extends JCheckBox implements ChangeListener {
    public int address;

    public BptButton () {
      address = 0;
      this.addChangeListener (this);
    }

    public void stateChanged (ChangeEvent ce) {
      if (isSelected ()) SparcSim.setbreakpoint (address);
                  else SparcSim.clearbreakpoint (address);
    }
  }
}
