//+++2004-11-01
//    Copyright (C) 2004  Mike Rieker, Beverly, MA USA
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//---2004-11-01

import java.io.*;
import java.util.*;
import javax.swing.*;

// These objects perform assembly of source modules
// To assemble some source files:
//  1) Create an SparcAsm instance to reset everything
//  2) Do pass 1 by calling assemble with each source file
//  3) Do pass 2 by calling assemble with each source file again

public class SparcAsm {
  public static int currentlocation;

  private static boolean pass2;
  private static SourceEditor sourceeditor;  // source editor containing source being assembled
  private static String sourcefilebuf;       // the whole source file being assembled
  private static String sourcelinebuf;       // the whole source line being assembled
  private static int sourcelinenum;          // line number in source file containing current line
  private static int sourcelineidx;          // index in sourcelinebuf for next token
  private static int sourcefileidx;          // character index in sourcefilebuf for next line
  private static boolean haveerrors;         // set to 'true' if any error detected in source file
  private static Sym dot;                    // points to '.' symbol table entry
  private static boolean popuperrorsenabled;

  private static final int TOK_ENDLINE =  1; // end of line
  private static final int TOK_SYMBOL  =  2; // symbol name (strvalue; intvalue=length)
  private static final int TOK_COLON   =  3; // colon
  private static final int TOK_COMMA   =  4; // comma
  private static final int TOK_OPBRACK =  5; // open bracket
  private static final int TOK_CLBRACK =  6; // close bracket
  private static final int TOK_ASSIGN  =  7; // = assignment equals
  private static final int TOK_QUOTED  =  8; // double quoted string (strvalue has string)
  private static final int TOK_INTEGER =  9; // integer constant
  private static final int TOK_MINUS   = 10; // - hyphen
  private static final int TOK_TILDE   = 11; // ~ tilde
  private static final int TOK_PERCENT = 12; // % percent
  private static final int TOK_AMPER   = 13; // & ampersand
  private static final int TOK_OPPAREN = 14; // ( open parenthesis
  private static final int TOK_CLPAREN = 15; // ) close parenthesis
  private static final int TOK_PLUS    = 16; // + plus
  private static final int TOK_SLASH   = 17; // / divide
  private static final int TOK_SPLAT   = 18; // * multiply
  private static final int TOK_BAR     = 19; // | or
  private static final int TOK_LSHIFT  = 20; // << left shift
  private static final int TOK_RSHIFT  = 21; // >> right shift

  private static final String valsymchars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_.$";

  private static class Op {
    int optype;
    int opcode;

    public Op (int t, int c) {
      optype = t;
      opcode = c;
    }
  }

  private static final int OP_ALIGN   =  1;
  private static final int OP_ARITH1A =  2;
  private static final int OP_ARITH1C =  3;
  private static final int OP_ARITH2A =  4;
  private static final int OP_ARITH2C =  5;
  private static final int OP_ARITH3  =  6;
  private static final int OP_ASCII   =  7;
  private static final int OP_BRANCH  =  8;
  private static final int OP_CALL    =  9;
  private static final int OP_DATA    = 10;
  private static final int OP_DEFINE  = 11;
  private static final int OP_FIXED   = 12;
  private static final int OP_GLOBAL  = 13;
  private static final int OP_JMP     = 14;
  private static final int OP_JMPL    = 15;
  private static final int OP_LOAD    = 16;
  private static final int OP_SETHI   = 17;
  private static final int OP_SKIP    = 18;
  private static final int OP_STORE   = 19;
  private static final int OP_MACRO   = 20;
  private static final int OP_INCL    = 21;

  private static HashMap ops = null;

  private static class Sym {
    boolean markedglobal;
    boolean isdefined;
    boolean fixed;
    int value;
  }

  private static HashMap globalsymbols;      // .global symbols, reset at beginning of all files
  private static HashMap localsymboltables;  // list indexed by SourceEditor
  private static HashMap localsymbols;       // current source file's local symbol table
  private static HashMap definitions;        // current source file's 'define' substitutions
  private static HashMap permasymbols;       // permanent symbols (via defpermasym) for IO device registers

  // Source files, nested via .include directives

  private static SrcFile srcfile;

  private static class SrcFile {
    public SrcFile outerfile;
    public String filename;
    public String currentline;
    public int linenumber;
    private BufferedReader sourcefile;

    public SrcFile (String n) {
      filename = n;
    }

    public boolean open () {
      try {
        sourcefile = new BufferedReader (new FileReader (filename));
      } catch (IOException ioe) {
        errormessage ("Error opening " + filename + ": " + ioe.getMessage ());
        return (false);
      }
      linenumber = 0;
      return (true);
    }

    public String readline () {
      try {
        currentline = sourcefile.readLine ();
        linenumber ++;
      } catch (IOException ioe) {
        errormessage ("Error reading " + filename + ": " + ioe.getMessage ());
        currentline = null;
      }
      return (currentline);
    }

    public void close () {
      try {
        sourcefile.close ();
      } catch (IOException ioe) {
      }
      sourcefile = null;
    }
  }

  // Macro definitions

  private static HashMap macrodefs;  // (Mac) macro definitions
                                     // key is the macro name string

  private static class Mac {
    public String name;              // this macro's name
    public ArrayList parameters;     // (String) array of parameter names
    public ArrayList codelines;      // (String) array of definition code lines

    public Mac (String n) {
      name = n;
      parameters = new ArrayList ();
      codelines  = new ArrayList ();
    }
  }

  // There is one of these for each macro currently being expanded

  private static MacExp macroexpand;   // macro being expanded (or null if none)

  private static class MacExp {
    public int lineindex;              // index of next line of macro to expand
    public String currentline;         // current line including expansions
    private Mac mac;                   // which macro is being expanded
    private MacExp nextouter;          // the macro we were called from (or null if original source file)
    private String[] arguments;        // the call argument values

    public MacExp (Mac m) {
      int i;

      mac = m;                         // save macro pointer being expanded
      lineindex = 0;                   // start with the first line

      // Fill in arguments array from the current source line

      arguments = new String[mac.parameters.size()];
      for (i = 0;;) {
        int begindex = sourcelineidx;
        int endindex = sourcelineidx;
        Tok tok;

        while (((tok = new Tok ()).tokentype != TOK_COMMA) && (tok.tokentype != TOK_ENDLINE)) {
          endindex = sourcelineidx;
        }
        if (i >= arguments.length) {
          errormessage ("Extra macro parameter " + sourcelinebuf.substring (begindex, endindex));
          break;
        }
        arguments[i++] = sourcelinebuf.substring (begindex, endindex);
        if (tok.tokentype == TOK_ENDLINE) break;
      }

      // If any missing, set them to null strings so we don't puque

      if (i < arguments.length) {
        errormessage ("Missing macro parameter " + (String) mac.parameters.get (i));
        while (i < arguments.length) arguments[i++] = "";
      }
    }

    // Retrieve next expanded line with substitutions made

    public String getnextline () {
      currentline = null;
      if (lineindex >= mac.codelines.size ()) {                     // see if reached the end of the macro's definition
        return (null);                                              // if so, tell caller we reached the end
      }
      StringBuffer codeline = new StringBuffer ((String) mac.codelines.get (lineindex ++));
      int cll = codeline.length ();                                 // get length of original mac def code line
      for (int i = 0; i < cll; i ++) {                              // scan thru line one char at a time
        if ((i > 0) && (valsymchars.indexOf (codeline.charAt (i - 1)) >= 0)) continue; // skip if predecessor is not a delimeter
        for (int j = 0; j < arguments.length; j ++) {               // scan thru argument list
          String p = (String) mac.parameters.get (j);               // get corresponding parameter name
          int l = p.length ();
          if (i + l > cll) continue;                                // make sure code line is long enuf for it
          if ((i + l < cll) && (valsymchars.indexOf (codeline.charAt (i + l)) >= 0)) continue; // skip if not followed by a delim
          if (p.equals (codeline.substring (i, i + l))) {           // see if it matches the parameter name
            String a = arguments[j];                                // if so, replace with argument string
            codeline.replace (i, i + l, (a));
            cll += a.length () - l;
            i += a.length () - 1;
            break;
          }
        }
      }
      currentline = codeline.toString ();
      return (currentline);
    }
  }

  // Define permanent symbols - never get wiped out by assembler

  public static void defpermasym (String name, int value) {
    if (permasymbols == null) permasymbols = new HashMap ();
    Sym sym = new Sym ();
    sym.markedglobal = true;
    sym.isdefined    = true;
    sym.fixed        = true;
    sym.value        = value;
    permasymbols.put (name, sym);
  }

  // Reset assembler in preparation for assembling a group of source files

  public static void reset () {

    // Reset the symbol tables

    globalsymbols = new HashMap ();      // key is symbol name string
                                         // value is a Sym

    localsymboltables = new HashMap ();  // key is source editor object
                                         // value is an HashMap of symbols for that source file
                                         //   key is symbol name string
                                         //   value is a Sym

    macrodefs = new HashMap ();          // key is macro name string
                                         // value is a Mac

    popuperrorsenabled = true;           // enable the popup error message dialog box

    // The opcode table only needs to be built once

    if (ops != null) return;
    ops = new HashMap ();

    ops.put (".align",   new Op (OP_ALIGN,  0));
    ops.put (".ascii",   new Op (OP_ASCII,  0));
    ops.put (".byte",    new Op (OP_DATA,   1));
    ops.put (".global",  new Op (OP_GLOBAL, 0));
    ops.put (".half",    new Op (OP_DATA,   2));
    ops.put (".include", new Op (OP_INCL,   0));
    ops.put (".macro",   new Op (OP_MACRO,  0));
    ops.put (".skip",    new Op (OP_SKIP,   0));
    ops.put (".word",    new Op (OP_DATA,   4));

    ops.put ("add",     new Op (OP_ARITH3,  0x80000000));
    ops.put ("addcc",   new Op (OP_ARITH3,  0x80800000));
    ops.put ("addx",    new Op (OP_ARITH3,  0x80400000));
    ops.put ("addxcc",  new Op (OP_ARITH3,  0x80C00000));
    ops.put ("and",     new Op (OP_ARITH3,  0x80080000));
    ops.put ("andcc",   new Op (OP_ARITH3,  0x80880000));
    ops.put ("andn",    new Op (OP_ARITH3,  0x80280000));
    ops.put ("andncc",  new Op (OP_ARITH3,  0x80A80000));
    ops.put ("b",       new Op (OP_BRANCH,  0x10800000));
    ops.put ("ba",      new Op (OP_BRANCH,  0x10800000));
    ops.put ("bcc",     new Op (OP_BRANCH,  0x1A800000));
    ops.put ("bcs",     new Op (OP_BRANCH,  0x0A800000));
    ops.put ("be",      new Op (OP_BRANCH,  0x02800000));
    ops.put ("bg",      new Op (OP_BRANCH,  0x14800000));
    ops.put ("bge",     new Op (OP_BRANCH,  0x16800000));
    ops.put ("bgeu",    new Op (OP_BRANCH,  0x1A800000));
    ops.put ("bgu",     new Op (OP_BRANCH,  0x18800000));
    ops.put ("bl",      new Op (OP_BRANCH,  0x06800000));
    ops.put ("ble",     new Op (OP_BRANCH,  0x04800000));
    ops.put ("bleu",    new Op (OP_BRANCH,  0x08800000));
    ops.put ("blu",     new Op (OP_BRANCH,  0x0A800000));
    ops.put ("bn",      new Op (OP_BRANCH,  0x00800000));
    ops.put ("bne",     new Op (OP_BRANCH,  0x12800000));
    ops.put ("bneg",    new Op (OP_BRANCH,  0x0C800000));
    ops.put ("bpos",    new Op (OP_BRANCH,  0x1C800000));
    ops.put ("bvc",     new Op (OP_BRANCH,  0x1E800000));
    ops.put ("bvs",     new Op (OP_BRANCH,  0x0E800000));
    ops.put ("call",    new Op (OP_CALL,    0x9FC00000)); // jmpl address,%o7
    ops.put ("cmp",     new Op (OP_ARITH2C, 0x80A00000));
    ops.put ("define",  new Op (OP_DEFINE,  0));
    ops.put ("jmp",     new Op (OP_JMP,     0x81C00000)); // jmpl address,%g0
    ops.put ("jmpl",    new Op (OP_JMPL,    0x81C00000));
    ops.put ("ld",      new Op (OP_LOAD,    0xC0000000));
    ops.put ("ldd",     new Op (OP_LOAD,    0xC0180000));
    ops.put ("ldsb",    new Op (OP_LOAD,    0xC0480000));
    ops.put ("ldsh",    new Op (OP_LOAD,    0xC0500000));
    ops.put ("ldub",    new Op (OP_LOAD,    0xC0080000));
    ops.put ("lduh",    new Op (OP_LOAD,    0xC0100000));
    ops.put ("mov",     new Op (OP_ARITH2A, 0x80000000));
    ops.put ("mulscc",  new Op (OP_ARITH3,  0x81200000));
    ops.put ("nop",     new Op (OP_FIXED,   0x80000000));
    ops.put ("or",      new Op (OP_ARITH3,  0x80100000));
    ops.put ("orcc",    new Op (OP_ARITH3,  0x80900000));
    ops.put ("orn",     new Op (OP_ARITH3,  0x80300000));
    ops.put ("orncc",   new Op (OP_ARITH3,  0x80B00000));
    ops.put ("rdy",     new Op (OP_ARITH1C, 0x81400000));
    ops.put ("restore", new Op (OP_FIXED,   0x81E80000));
    ops.put ("ret",     new Op (OP_FIXED,   0x81C7E008)); // jmpl %i7+8,%g0
    ops.put ("retl",    new Op (OP_FIXED,   0x81C3E008)); // jmpl %o7+8,%g0
    ops.put ("save",    new Op (OP_ARITH3,  0x81E00000));
    ops.put ("sethi",   new Op (OP_SETHI,   0x01000000));
    ops.put ("sll",     new Op (OP_ARITH3,  0x81280000));
    ops.put ("sra",     new Op (OP_ARITH3,  0x81380000));
    ops.put ("srl",     new Op (OP_ARITH3,  0x81300000));
    ops.put ("st",      new Op (OP_STORE,   0xC0200000));
    ops.put ("stb",     new Op (OP_STORE,   0xC0280000));
    ops.put ("std",     new Op (OP_STORE,   0xC0380000));
    ops.put ("sth",     new Op (OP_STORE,   0xC0300000));
    ops.put ("sub",     new Op (OP_ARITH3,  0x80200000));
    ops.put ("subcc",   new Op (OP_ARITH3,  0x80A00000));
    ops.put ("subx",    new Op (OP_ARITH3,  0x80600000));
    ops.put ("subxcc",  new Op (OP_ARITH3,  0x80E00000));
    ops.put ("swap",    new Op (OP_LOAD,    0xC0780000));
    ops.put ("tst",     new Op (OP_ARITH1A, 0x80900000)); // orcc reg,%g0,%g0
    ops.put ("wry",     new Op (OP_ARITH2C, 0x81800000));
    ops.put ("xnor",    new Op (OP_ARITH3,  0x80380000));
    ops.put ("xnorcc",  new Op (OP_ARITH3,  0x80B80000));
    ops.put ("xor",     new Op (OP_ARITH3,  0x80180000));
    ops.put ("xorcc",   new Op (OP_ARITH3,  0x80980000));
  }

  //  Assemble a source file and put results in memory
  //    Input:
  //      se = source file editor
  //      p2 = false : this is pass one (generate symbol table only)
  //            true : this is pass two (output object code)
  //      loc = starting memory address
  //    Output:
  //      assemble = true : successful
  //                false : failed (messages already displayed)
  //      currentlocation = updated to point to next available location

  public static boolean assemble (SourceEditor se, boolean p2, int loc) {
    currentlocation = loc;

    sourceeditor  = se;
    sourcefilebuf = sourceeditor.gettextstring ();        // get whole file as a single string
    sourcelinenum = 0;                                    // we're about to get first line
    sourcefileidx = 0;                                    // it starts at beg of file

    pass2 = p2;
    haveerrors = false;
    macroexpand = null;

    localsymbols = (HashMap) localsymboltables.get (se);  // get local symbol table pointer
    if (localsymbols == null) {
      dot = new Sym ();
      dot.fixed = false;
      dot.isdefined = true;
      dot.markedglobal = false;
      localsymbols = new HashMap ();
      localsymbols.put (".", dot);                        // make sure it has a "." symbol
      localsymboltables.put (se, localsymbols);
    } else {
      dot = (Sym) localsymbols.get (".");
    }
    dot.value = loc;                                      // set "." to starting mem location

    definitions = new HashMap ();                         // clear/init "define" definitions

    sourceeditor.resetasminfo ();                         // clear asm info from source editor window

    while (fetchsourceline ()) {
      processsourceline ();
    }
    definitions = null;

    // Move any symbols in the local table that are marked global to the global table

    Iterator localsymboliterator = localsymbols.keySet ().iterator ();  // set up to loop through local table
    while (localsymboliterator.hasNext ()) {                            // loop while there's stuff to do
      String name = (String) localsymboliterator.next ();               // get name of next local symbol
      Sym lclsym = (Sym) localsymbols.get (name);                       // get its value
      if (lclsym.isdefined && lclsym.markedglobal) {                    // see if defined and marked global
        Sym gblsym = (Sym) globalsymbols.get (name);                    // ok, see if already have a global by that name
        if (gblsym == null) globalsymbols.put (name, lclsym);           //     if not, define one
        else if (gblsym.value != lclsym.value) {                        //     else, the value better match
          errormessage (name + " definition conflict");
        }
      }
    }
    localsymbols = null;                                                // don't need pointer to table any more

    currentlocation = dot.value;
    dot = null;

    return (!haveerrors);
  }

  // Extract next line from source file
  //  Input:
  //   macroexpand = macro being expanded (or null if none)
  //   srcfile = .include source file (or null if none)
  //   sourcefilebuf = whole original source file as displayed in window
  //   sourcefileidx = index of line to fetch from sourcefilebuf
  //  Output:
  //   fetchsourceline = false : end of file
  //                      true : source line fetched
  //   sourcelinebuf = source line string (not including line terminators)
  //   sourcelineidx = 0
  //   mcrosexpand,srcfile,or sourcefileidx = incremented for next line to fetch
  //   sourcelinenum = incremented (line number)

  // Yes, now ask if I *care* that I didn't implement polymorphism for the three sources of source

  private static boolean fetchsourceline () {

    sourcelinebuf = null;

    // Check for macro expansion in progress
    // If so, get the next macro line and process it

    while (macroexpand != null) {
      sourcelinebuf = macroexpand.getnextline ();
      if (sourcelinebuf != null) break;
      macroexpand = macroexpand.nextouter;
    }

    // If we didn't get it from macro, try reading next include file line

    if (sourcelinebuf == null) {
      while (srcfile != null) {
        sourcelinebuf = srcfile.readline ();
        if (sourcelinebuf != null) break;
        srcfile.close ();
        srcfile = srcfile.outerfile;
      }
    }

    // If no source file, try getting from original file in the window

    if (sourcelinebuf == null) {
      int i = sourcefilebuf.indexOf ('\n', sourcefileidx);         // find end of current line
      if (i < 0) i = sourcefilebuf.length ();                      // make sure we get the last bit
      if (i < sourcefileidx) return (false);                       // tell caller we're all done
      sourcelinebuf = sourcefilebuf.substring (sourcefileidx, i);  // get the line
      sourcefileidx = ++ i;                                        // remember where next one starts
      sourcelinenum ++;                                            // increment line number
    }

    // Start parsing at beginning of line

    sourcelineidx = 0;
    return (true);
  }

  // Process 'sourcelinebuf'

  private static void processsourceline () {
    Tok nextok, tok;

    // Parse any <label>: at beginning of line

    nextok = null;
    while ((tok = new Tok ()).tokentype != TOK_ENDLINE) {

      // Line must begin with a symbol (either a label or an opcode)

      if (tok.tokentype != TOK_SYMBOL) {
        errormessage ("Bad token %s at beg of line" + tok.strvalue);
        return;
      }

      // If next token on line is a colon, the symbol is a label
      // Note that we may have more than one on a line

      nextok = new Tok();
      if (nextok.tokentype == TOK_COLON) {
        Sym sym;
        sym = (Sym) localsymbols.get (tok.strvalue);
        if (sym == null) {
          sym = new Sym ();
          sym.fixed = true;
          sym.isdefined = true;
          sym.markedglobal = false;
          sym.value = dot.value;
          localsymbols.put (tok.strvalue, sym);
        } else if (!sym.isdefined) {
          sym.fixed = true;
          sym.isdefined = true;
          sym.value = dot.value;
        } else if (!sym.fixed || (sym.value != dot.value)) {
          errormessage ("Symbol multiple definition");
        }
        continue;
      }
      break;
    }
    if (tok.tokentype == TOK_ENDLINE) return;

    // If next token on line is an '=', evaluate expression and assign to symbol

    if (nextok.tokentype == TOK_ASSIGN) {
      Sym sym;
      sym = (Sym) localsymbols.get (tok.strvalue);
      if (sym == null) {
        sym = new Sym ();
        sym.fixed = false;
        sym.isdefined = true;
        sym.markedglobal = false;
        sym.value = evaluate (TOK_ENDLINE);
        localsymbols.put (tok.strvalue, sym);
      } else {
        if (sym.fixed) errormessage ("Fixed symbol cannot be redefined");
        else sym.value = evaluate (TOK_ENDLINE);
        sym.isdefined = true;
      }
      return;
    }

    // Undo what might have been the colon or the equal

    nextok.undo ();

    // Well, maybe it's really a macro expansion

    Mac mac = (Mac) macrodefs.get (tok.strvalue);
    if (mac != null) {
      MacExp macexp = new MacExp (mac);
      macexp.nextouter = macroexpand;
      macroexpand = macexp;
      return;
    }

    // OK, it better be an opcode or directive

    Op op = (Op) ops.get (tok.strvalue.toLowerCase ());
    if (op == null) {
      errormessage ("Bad opcode/directive/macro " + tok.strvalue);
      return;
    }

    switch (op.optype) {

      // .align 1,2,4,8 or 16

      case OP_ALIGN: {
        int alignto = evaluate (TOK_ENDLINE);
        if ((alignto == 1) || (alignto == 2) || (alignto == 4) || (alignto == 8) || (alignto == 16)) {
          dot.value = (dot.value + alignto - 1) & - alignto;
        } else {
          errormessage ("Alignment must be 1,2,4,8,16");
        }
        break;
      }

      // One-operand arithmetic: <opcode> <reg>,%g0,%g0

      case OP_ARITH1A: {
        int reg = getregnum (TOK_ENDLINE);
        output32 (op.opcode + (reg << 14));
        break;
      }

      // One-operand arithmetic: <opcode> %g0,%g0,<reg>

      case OP_ARITH1C: {
        int reg = getregnum (TOK_ENDLINE);
        output32 (op.opcode + (reg << 25));
        break;
      }

      // Two-operand arithmetic: <opcode> %g0,<ri2>,<rd>

      case OP_ARITH2A: {
        int ri2 = getregimm (TOK_COMMA);
        int rd  = getregnum (TOK_ENDLINE);
        output32 (op.opcode + ri2 + (rd << 25));
        break;
      }

      // Two-operand arithmetic: <opcode> <rs1>,<ri2>,%g0

      case OP_ARITH2C: {
        int rs1 = getregnum (TOK_COMMA);
        int ri2 = getregimm (TOK_ENDLINE);
        output32 (op.opcode + (rs1 << 14) + ri2);
        break;
      }

      // Three-operand arithmetic: <opcode> <rs1>,<ri2>,<rd>

      case OP_ARITH3: {
        int rs1 = getregnum (TOK_COMMA);
        int ri2 = getregimm (TOK_COMMA);
        int rd  = getregnum (TOK_ENDLINE);
        output32 (op.opcode + (rs1 << 14) + ri2 + (rd << 25));
        break;
      }

      // ASCII strings in double quotes

      case OP_ASCII: {
        while (true) {
          tok = new Tok ();
          if (tok.tokentype == TOK_ENDLINE) break;
          if (tok.tokentype == TOK_QUOTED) {
            for (int j = 0; j < tok.strvalue.length (); j ++) {
              output8 (tok.strvalue.charAt (j));
            }
          } else if (tok.tokentype == TOK_INTEGER) {
            output8 (tok.intvalue);
          } else if (tok.tokentype != TOK_COMMA) {
            errormessage ("Bad string token " + nextok.strvalue);
            break;
          }
        }
        break;
      }

      // Branch: <opcode>[,a] <destination>

      case OP_BRANCH: {
        int abit;
        int dest;
        Tok comma;

        comma = new Tok ();
        abit = 0;
        if (comma.tokentype == TOK_COMMA) {
          comma = new Tok ();
          if ((comma.tokentype == TOK_SYMBOL) && comma.strvalue.equalsIgnoreCase ("a")) {
            abit = 0x20000000;
          } else {
            errormessage ("Branch comma not followed by 'a' annul flag");
          }
        } else {
          comma.undo ();
        }
        dest   = evaluate (TOK_ENDLINE);
        dest  -= dot.value;
        if ((dest & 3) != 0) {
          errormessage ("Unaligned branch destination");
        }
        dest >>= 2;
        dest  &= 0x3FFFFF;
        output32 (op.opcode + abit + dest);
        break;
      }

      // Call: <opcode> <destination>

      case OP_CALL: {
        Tok tok2 = new Tok ();
        tok2.undo ();
        if (tok2.tokentype == TOK_PERCENT) {
          int adr = getaddrexpnb (TOK_COMMA);
          output32 (op.opcode + adr);
        } else {
          int dest = evaluate (TOK_ENDLINE);
          dest -= dot.value;
          if ((dest & 3) != 0) {
            errormessage ("Unaligned call destination");
          }
          dest >>= 2;
          dest &= 0x3FFFFFFF;
          output32 (0x40000000 + dest);
        }
        break;
      }

      // Data definition: <size> <value>,...

      case OP_DATA: {
        do {
          int value = evaluate (0);
          nextok = null;
          switch (op.opcode) {
            case 1: output8  (value); break;
            case 2: output16 (value); break;
            case 4: output32 (value); break;
          }
          tok = new Tok ();
        } while (tok.tokentype == TOK_COMMA);
        if (tok.tokentype != TOK_ENDLINE) {
          errormessage ("Bad data value " + tok.strvalue);
        }
        break;
      }

      // DEFINE (symbol, token)

      case OP_DEFINE: {
        tok = new Tok ();
        if (tok.tokentype != TOK_OPPAREN) errormessage ("Missing ( after define");
        else {
          tok = new Tok (false);
          if (tok.tokentype != TOK_SYMBOL) errormessage ("Missing symbolname after (");
          else {
            String symbol = tok.strvalue;
            tok = new Tok ();
            if (tok.tokentype != TOK_COMMA) errormessage ("Missing , after symbolname");
            else {
              tok = new Tok (false);
              if ((tok.tokentype != TOK_INTEGER) && (tok.tokentype != TOK_SYMBOL)) {
                errormessage ("Missing integer or symbolname after ,");
              } else {
                Tok token = tok;
                tok = new Tok ();
                if (tok.tokentype != TOK_CLPAREN) errormessage ("Missing ) at end");
                else {
                  tok = new Tok ();
                  if (tok.tokentype != TOK_ENDLINE) errormessage ("Extra stuff after )");
                  else {
                    definitions.put (symbol, token);
                  }
                }
              }
            }
          }
        }
        break;
      }

      // Fixed: <opcode> (no operands)

      case OP_FIXED: {
        tok = new Tok ();
        if (tok.tokentype != TOK_ENDLINE) {
          errormessage ("Extra stuff at end of line");
        }
        output32 (op.opcode);
        break;
      }

      // .GLOBAL <symbol>,...

      case OP_GLOBAL: {
        while ((tok = new Tok ()).tokentype != TOK_ENDLINE) {
          if (tok.tokentype == TOK_COMMA) continue;
          if (tok.tokentype != TOK_SYMBOL) errormessage ("bad symbol name " + tok.strvalue);
          else {
            Sym sym = (Sym) localsymbols.get (tok.strvalue);
            if (sym != null) sym.markedglobal = true;
            else {
              sym = new Sym ();
              sym.isdefined = false;
              sym.markedglobal = true;
              localsymbols.put (tok.strvalue, sym);
            }
          }
        }
        break;
      }

      // .INCLUDE "filename"

      case OP_INCL: {
        tok = new Tok ();
        if (tok.tokentype != TOK_QUOTED) {
          errormessage ("Included filename must be in double quotes");
        } else {
          SrcFile newsrcfile = new SrcFile (tok.strvalue);
          if (newsrcfile.open ()) {
            newsrcfile.outerfile = srcfile;
            srcfile = newsrcfile;
          }
        }
        break;
      }

      // JMP <address>,%g0

      case OP_JMP: {
        int adr = getaddrexpnb (TOK_ENDLINE);
        output32 (op.opcode + adr);
        break;
      }

      // JMPL <address>,<register>

      case OP_JMPL: {
        int adr = getaddrexpnb (TOK_COMMA);
        int rd  = getregnum (TOK_ENDLINE);
        output32 (op.opcode + (rd << 25) + adr);
        break;
      }

      // Load: <opcode> <address>,<register>

      case OP_LOAD: {
        int adr = getaddrexp (TOK_COMMA);
        int rd  = getregnum (TOK_ENDLINE);
        output32 (op.opcode + (rd << 25) + adr);
        break;
      }

      // .MACRO <name> <parameter>,...

      case OP_MACRO: {

        // Process the .MACRO directive line itself

        Tok nametok = new Tok ();
        if (nametok.tokentype != TOK_SYMBOL) {
          errormessage ("Bad or missing macro name");
          break;
        }
        mac = new Mac (nametok.strvalue);
        while ((tok = new Tok ()).tokentype == TOK_SYMBOL) {
          mac.parameters.add (tok.strvalue);
          tok = new Tok ();
          if (tok.tokentype != TOK_COMMA)break;
        }
        if (tok.tokentype != TOK_ENDLINE) {
          errormessage ("Bad macro parameter token " + tok.strvalue);
        }

        // Now schlep up source lines until a matching '.ENDM <name>' is found

        while (true) {
          if (!fetchsourceline ()) {
            errormessage ("End-of-file during macro definition " + nametok.strvalue);
            break;
          }
          tok = new Tok ();
          if ((tok.tokentype == TOK_SYMBOL) && tok.strvalue.equalsIgnoreCase (".endm")) {
            macrodefs.put (nametok.strvalue, mac);
            break;
          }
          mac.codelines.add (sourcelinebuf);
        }
        break;
      }

      // SETHI <value>,<register>

      case OP_SETHI: {
        int value = evaluate (TOK_COMMA) & 0x3FFFFF;
        int rd = getregnum (TOK_ENDLINE);
        output32 (op.opcode + (rd << 25) + value);
        break;
      }

      // .skip <numberofbytes>

      case OP_SKIP: {
        int nbytes = evaluate (TOK_ENDLINE);
        dot.value += nbytes;
        break;
      }

      // Store: <opcode> <register>,<address>

      case OP_STORE: {
        int rs  = getregnum (TOK_COMMA);
        int adr = getaddrexp (TOK_ENDLINE);
        output32 (op.opcode + (rs << 25) + adr);
        break;
      }
    }
  }

  // Get the start address (from symbol _start)

  public static int getstartaddr () {
    Sym startsym;

    if (globalsymbols != null) {
      startsym = (Sym) globalsymbols.get ("_start");
      if ((startsym != null) && startsym.isdefined) return (startsym.value);
      JOptionPane.showMessageDialog (null, "No global symbol _start defined");
    }
    return (0);
  }

  //  Get address expression

  private static int getaddrexp (int terminator) {
    Tok tok2 = new Tok ();
    if (tok2.tokentype != TOK_OPBRACK) {
      errormessage ("Missing [ for address expression");
    }
    int rs1 = getregnum (0);
    int ri2 = 0;
    tok2 = new Tok ();
    if ((tok2.tokentype == TOK_PLUS) || (tok2.tokentype == TOK_MINUS)) {
      if (tok2.tokentype == TOK_MINUS) tok2.undo ();
      ri2 = getregimm (TOK_CLBRACK);
    } else if (tok2.tokentype != TOK_CLBRACK) {
      errormessage ("Missing ] or +-regimm] after load register");
    }
    tok2 = new Tok ();
    if (tok2.tokentype != terminator) {
      errormessage ("Bad address expression terminator " + tok2.strvalue);
    }
    return ((rs1 << 14) + ri2);
  }

  // Get address expression without brackets (for JMPL)

  private static int getaddrexpnb (int terminator) {
    int rs1 = getregnum (0);
    int ri2 = 0;
    Tok tok2 = new Tok ();
    if ((tok2.tokentype == TOK_PLUS) || (tok2.tokentype == TOK_MINUS)) {
      if (tok2.tokentype == TOK_MINUS) tok2.undo ();
      ri2 = getregimm (TOK_CLBRACK);
    } else {
      tok2.undo ();
    }
    tok2 = new Tok ();
    if (tok2.tokentype != terminator) {
      errormessage ("Bad address expression terminator " + tok2.strvalue);
    }
    return ((rs1 << 14) + ri2);
  }

  //  Get register number or immediate value from line
  //    Input:
  //      terminator = TOK_whatever : must have this terminator next
  //    Output:
  //      getregimm = value encoded for arith instruction

  private static int getregimm ( int terminator) {
    int retval;
    Tok tok;

    tok = new Tok ();
    if (tok.tokentype == TOK_PERCENT) {
      Tok nextok = new Tok ();
      if ((nextok.tokentype != TOK_SYMBOL)
       || (!nextok.strvalue.equalsIgnoreCase ("hi")
        && !nextok.strvalue.equalsIgnoreCase ("lo"))) {
        tok.undo ();
        return (getregnum (terminator));
      }
    }

    tok.undo ();
    retval = evaluate (terminator);
    if ((retval < -4096) || (retval > 4095)) {
      errormessage ("Immediate value " + retval + " out of range");
    }
    return ((retval & 0x1FFF) | 0x2000);
  }

  //  Get register number token from line
  //    Input:
  //      terminator = TOK_whatever : must have this terminator next
  //                           or 0 : don't check terminator
  //    Output:
  //      getregnum = register number 0..31

  private static int getregnum (int terminator) {
    int cc, regnum;
    Tok termtok, tok;

    final int[] reglist = { (0 << 16) + ('g' << 8) + '0',  (16 << 16) + ('l' << 8) + '0',
                            (1 << 16) + ('g' << 8) + '1',  (17 << 16) + ('l' << 8) + '1',
                            (2 << 16) + ('g' << 8) + '2',  (18 << 16) + ('l' << 8) + '2',
                            (3 << 16) + ('g' << 8) + '3',  (19 << 16) + ('l' << 8) + '3',
                            (4 << 16) + ('g' << 8) + '4',  (20 << 16) + ('l' << 8) + '4',
                            (5 << 16) + ('g' << 8) + '5',  (21 << 16) + ('l' << 8) + '5',
                            (6 << 16) + ('g' << 8) + '6',  (22 << 16) + ('l' << 8) + '6',
                            (7 << 16) + ('g' << 8) + '7',  (23 << 16) + ('l' << 8) + '7',
                            (8 << 16) + ('o' << 8) + '0',  (24 << 16) + ('i' << 8) + '0',
                            (9 << 16) + ('o' << 8) + '1',  (25 << 16) + ('i' << 8) + '1',
                           (10 << 16) + ('o' << 8) + '2',  (26 << 16) + ('i' << 8) + '2',
                           (11 << 16) + ('o' << 8) + '3',  (27 << 16) + ('i' << 8) + '3',
                           (12 << 16) + ('o' << 8) + '4',  (28 << 16) + ('i' << 8) + '4',
                           (13 << 16) + ('o' << 8) + '5',  (29 << 16) + ('i' << 8) + '5',
                           (14 << 16) + ('o' << 8) + '6',  (30 << 16) + ('i' << 8) + '6',
                           (15 << 16) + ('o' << 8) + '7',  (31 << 16) + ('i' << 8) + '7',
                           (14 << 16) + ('s' << 8) + 'p',  (30 << 16) + ('f' << 8) + 'p' };

    tok = new Tok ();
    if (tok.tokentype != TOK_PERCENT) {
      errormessage ("Expected register at " + tok.strvalue);
      return (0);
    }
    tok = new Tok ();
    if ((tok.tokentype != TOK_SYMBOL) || (tok.intvalue != 2)) {
      errormessage ("Expected register at " + tok.strvalue);
      return (0);
    }
    cc  = tok.strvalue.charAt (0);
    if ((cc >= 'A') && (cc <= 'Z')) cc += 'a' - 'A';
    cc <<= 8;
    cc += tok.strvalue.charAt (1);
    if (((cc & 0xFF) >= 'A') && ((cc & 0xFF) <= 'Z')) cc += 'a' - 'A';
    for (regnum = reglist.length; -- regnum >= 0;) {
      if ((reglist[regnum] & 0xFFFF) == cc) break;
    }
    if (regnum < 0) {
      errormessage ("Bad register name " + tok.strvalue);
      return (0);
    }
    regnum = reglist[regnum] >> 16;
    if (terminator != 0) {
      termtok = new Tok ();
      if (termtok.tokentype != terminator) {
        errormessage ("Invalid terminator following register at " + tok.strvalue);
        return (0);
      }
    }
    return (regnum);
  }

  // Evaluate a numeric expression
  //  Input:
  //   source pointer = first token in expression
  //   terminator = 0 : don't look for any particular terminator token type
  //             else : terminator must be this token type (comma, endline, etc)
  //  Output:
  //   evaluate = value of expression
  //   terminator = 0 : source points at terminating token
  //             else : source points past terminating token

  private static int evaluate (int terminator) {
    int left;
    Tok optok;

    // Get left-hand operand

    left = getoperand ();

    while (true) {

      // Get binary operator

      optok = new Tok ();
      if (optok.tokentype == terminator) return (left);

      // Process it

      switch (optok.tokentype) {
        case TOK_PLUS: {
          left += getoperand ();
          break;
        }
        case TOK_MINUS: {
          left -= getoperand ();
          break;
        }
        case TOK_PERCENT: {
          int divisor = getoperand ();
          if (divisor == 0) errormessage ("Divide by zero");
          else left %= getoperand ();
          break;
        }
        case TOK_AMPER: {
          left &= getoperand ();
          break;
        }
        case TOK_SLASH: {
          int divisor = getoperand ();
          if (divisor == 0) errormessage ("Divide by zero");
          else left /= divisor;
          break;
        }
        case TOK_SPLAT: {
          left *= getoperand ();
          break;
        }
        case TOK_LSHIFT: {
          left <<= getoperand ();
          break;
        }
        case TOK_RSHIFT: {
          left >>= getoperand ();
          break;
        }
        default: {
          if (terminator == 0) optok.undo ();
          else {
            errormessage ("Invalid operator " + optok.strvalue);
            do optok = new Tok ();
            while (optok.tokentype != terminator);
          }
          return (left);
        }
      }
    }
  }

  // Get evaluation operand from input stream
  //  Input:
  //   source pointer = first token in expression
  //  Output:
  //   getoperand = operand's value
  //   source pointer = just past operand token(s)

  private static int getoperand () {
    Tok tok;

    tok = new Tok ();
    switch (tok.tokentype) {

      // Open parenthesis - evaluate enclosed expression
      // It must terminate on a close parenthesis

      case TOK_OPPAREN: {
        return (evaluate (TOK_CLPAREN));
      }

      // Integer - that's the operand's value

      case TOK_INTEGER: {
        return (tok.intvalue);
      }

      // Symbol - the operand's value is the symbol's value

      case TOK_SYMBOL: {
        Sym sym;

        sym = (Sym) localsymbols.get (tok.strvalue);
        if ((sym == null) || !sym.isdefined) sym = (Sym) globalsymbols.get (tok.strvalue);
        if ((sym == null) || !sym.isdefined) {
          if (permasymbols == null) permasymbols = new HashMap ();
          sym = (Sym) permasymbols.get (tok.strvalue);
        }
        if ((sym != null) && sym.isdefined) return (sym.value);
        errormessage ("Undefined symbol " + tok.strvalue);
        return (0);
      }

      // Hyphen - the operand's value is the negative of following operand

      case TOK_MINUS: {
        return (- getoperand ());
      }

      // Tilde - the operand's value is the complement of following operand

      case TOK_TILDE: {
        return (~ getoperand ());
      }

      // Maybe %hi or %lo

      case TOK_PERCENT: {
        Tok nextok = new Tok ();
        if (nextok.tokentype == TOK_SYMBOL) {

          // %hi means return the value of subsequent operand >> 10

          if (nextok.strvalue.equalsIgnoreCase ("hi")) {
            return (getoperand () >> 10);
          }

          // %lo means return the value of subsequent operand & 0x3FF

          if (nextok.strvalue.equalsIgnoreCase ("lo")) {
            return (getoperand () & 0x3FF);
          }
        }

        // Undo the symbol (or whatever) token and fall through to error message

        nextok.undo ();
      }

      // Don't know what the operand is

      default: {
        errormessage ("Invalid operand " + tok.strvalue);
        return (0);
      }
    }
  }

  // Source input tokens

  private static class Tok {
    private final int[] dchrlist = { (TOK_LSHIFT << 16) + ('<' << 8) + '<',
                                     (TOK_RSHIFT << 16) + ('>' << 8) + '>' };

    private final int[] schrlist = { (TOK_COLON   << 8) + ':',
                                     (TOK_COMMA   << 8) + ',',
                                     (TOK_OPBRACK << 8) + '[',
                                     (TOK_CLBRACK << 8) + ']',
                                     (TOK_ASSIGN  << 8) + '=',
                                     (TOK_OPPAREN << 8) + '(',
                                     (TOK_CLPAREN << 8) + ')',
                                     (TOK_MINUS   << 8) + '-',
                                     (TOK_TILDE   << 8) + '~',
                                     (TOK_PLUS    << 8) + '+',
                                     (TOK_PERCENT << 8) + '%',
                                     (TOK_AMPER   << 8) + '&',
                                     (TOK_SLASH   << 8) + '/',
                                     (TOK_SPLAT   << 8) + '*',
                                     (TOK_BAR     << 8) + '|' };

    public int tokentype;
    public String strvalue;
    public int intvalue;

    private int srcindex;

    // Constructor - parses the next token from input stream
    //  Input:
    //   substitute = false : don't process any 'define' substitutions
    //                 true : process 'define' substitutions
    //   source pointer = points to token to be parsed
    //  Output:
    //   source pointer = advanced past token just parsed

    public Tok () {
      this (true);
    }

    public Tok (boolean substitute) {
      char c;
      int cc, i, sbl;

      srcindex = sourcelineidx;

      i = 0;
      sbl = sourcelinebuf.length ();
      while (sourcelineidx < sbl) {
        c = sourcelinebuf.charAt (sourcelineidx);

        // Ignore whitespace

        if (c <= ' ') {
          sourcelineidx ++;
          continue;
        }

        // Exclamation mark is a comment and we flush the rest of the line

        if (c == '!') {
          sourcelineidx  = sourcelinebuf.length ();
          this.tokentype = TOK_ENDLINE;
          return;
        }

        // Double quote means a long string

        if (c == '\"') {
          sourcelineidx ++;
          this.tokentype = TOK_QUOTED;
          this.strvalue  = getquotedstring ('\"');
          this.intvalue  = this.strvalue.length ();
          return;
        }

        // Single quote means a string that is an integer of ASCII equivalent

        if (c == '\'') {
          String sb;

          this.tokentype = TOK_INTEGER;
          i = sourcelineidx ++;
          sb = getquotedstring ('\'');
          this.intvalue = 0;
          this.strvalue = sourcelinebuf.substring (i, sourcelineidx);
          if (sb.length () > 4) {
            errormessage("Single-quoted string too long " + sb);
          } else {
            for (i = 0; i < sb.length (); i ++) {
              this.intvalue = (this.intvalue << 8) + sb.charAt (i);
            }
          }
          return;
        }

        // If first character is a digit, then it's a numeric constant

        if ((c >= '0') && (c <= '9')) {
          int base;

          this.tokentype = TOK_INTEGER;
          this.intvalue = 0;
          base = 10;
          if (c == '0') {
            base = 8;
            if (((sourcelineidx + 1) < sbl) && ((sourcelinebuf.charAt (sourcelineidx + 1) == 'x')
                                             || (sourcelinebuf.charAt (sourcelineidx + 1) == 'X'))) {
              c = 'x';
              for (i = 2; i + sourcelineidx < sbl; i ++) {
                c = sourcelinebuf.charAt (sourcelineidx + i);
                if ((c >= '0') && (c <= '9')) this.intvalue = (this.intvalue * 16) + c - '0';
                else if ((c >= 'A') && (c <= 'F')) this.intvalue = (this.intvalue * 16) + c - 'A' + 10;
                else if ((c >= 'a') && (c <= 'f')) this.intvalue = (this.intvalue * 16) + c - 'a' + 10;
                else break;
                c = 0;
              }
              base = 0;
            }
          }
          if (base > 0) {
            for (i = 0; i + sourcelineidx < sbl; i ++) {
              c = sourcelinebuf.charAt (sourcelineidx + i);
              if ((c < '0') || (c >= '0' + base)) break;
              this.intvalue = (this.intvalue * base) + c - '0';
              c = 0;
            }
          }
          if (valsymchars.indexOf (c) >= 0) {
            errormessage ("Illegal digit in " + sourcelinebuf.substring (sourcelineidx, sourcelineidx + i + 1));
          }
          this.strvalue = sourcelinebuf.substring (sourcelineidx, sourcelineidx + i);
          sourcelineidx += i;
          return;
        }

        // If first char is valid symbol char (other than a digit), it's a symbol name

        if (valsymchars.indexOf (c) >= 0) {

          // Gather up the symbol name into strvalue

          this.tokentype = TOK_SYMBOL;
          for (i = 1; i + sourcelineidx < sbl; i ++) {
            c = sourcelinebuf.charAt (sourcelineidx + i);
            if (valsymchars.indexOf (c) < 0) break;
          }
          this.intvalue = i;
          this.strvalue = sourcelinebuf.substring (sourcelineidx, sourcelineidx + i);
          sourcelineidx += i;

          // If substitutions enabled and it's in the definitions list, get the substitute token

          if (substitute) {
            Tok def = (Tok) definitions.get (this.strvalue);
            if (def != null) {
              this.tokentype = def.tokentype;
              this.strvalue  = def.strvalue;
              this.intvalue  = def.intvalue;
            }
          }
          return;
        }

        // Check for double-character tokens

        if (sourcelineidx + 1 < sbl) {
          cc = (((int)c) << 8) + (int)sourcelinebuf.charAt (sourcelineidx + 1);
          for (i = dchrlist.length; -- i >= 0;) {
            if ((dchrlist[i] & 0xFFFF) == cc) break;
          }
          if (i >= 0) {
            this.tokentype = dchrlist[i] >> 16;
            this.strvalue = sourcelinebuf.substring (sourcelineidx, sourcelineidx + 2);
            sourcelineidx += 2;
            return;
          }
        }

        // Check for single-character tokens

        for (i = schrlist.length; -- i >= 0;) {
          if ((schrlist[i] & 0xFF) == c) break;
        }
        if (i >= 0) {
          this.tokentype = schrlist[i] >> 8;
          this.strvalue = sourcelinebuf.substring (sourcelineidx, sourcelineidx + 1);
          sourcelineidx ++;
          return;
        }

        // Who knows?

        errormessage ("Invalid character " + c);
        sourcelineidx ++;
      }

      // No more input, it's the end of line

      this.tokentype = TOK_ENDLINE;
    }

    // Undo the token, ie, restore source file as it was before token was parsed

    private void undo () {
      sourcelineidx = this.srcindex;
    }
  }

  // Get a quoted string from the source line
  //  Input:
  //   terminator = single or double quote character
  //   sourceindex = points past the initial quote character
  //  Output:
  //   getquotedstring = string with all \'s processed
  //   sourceindex = incremented past termination character

  private static String getquotedstring (char terminator) {
    char c;
    StringBuffer sb;

    sb = new StringBuffer ();
    while (sourcelineidx < sourcelinebuf.length ()) {
      c = sourcelinebuf.charAt (sourcelineidx ++);
      if (c == terminator) return (sb.toString ());
      if (c == '\\') {
        if (sourcelineidx >= sourcelinebuf.length ()) break;
        c = sourcelinebuf.charAt (sourcelineidx ++);
        if (c == 'b')  c = '\b';
        if (c == 'n')  c = '\n';
        if (c == 't')  c = '\t';
        if (c == '\\') c = '\\';
        if (c == '\'') c = '\'';
        if (c == '\"') c = '\"';
      }
      sb.append (c);
    }
    errormessage ("End-of-line in quoted string");
    return (sb.toString ());
  }

  // Write output to memory
  // Also send it to source editor so it can be displayed with source line

  private static void output32 (int value) {
    if ((dot.value & 3) != 0) {
      errormessage ("Writing 32-bit unaligned value at " + HexTextField.hexstring (dot.value));
    }
    else if (pass2) {
      BigEndianMemory.write32 (dot.value, value);
      sourceeditor.objcode (sourcelinenum, dot.value, 4, value);
    }
    dot.value += 4;
  }

  private static void output16 (int value) {
    if ((dot.value & 1) != 0) {
      errormessage ("Writing 16-bit unaligned value at " + HexTextField.hexstring (dot.value));
    }
    else if (pass2) {
      BigEndianMemory.write16 (dot.value, value);
      sourceeditor.objcode (sourcelinenum, dot.value, 2, value);
    }
    dot.value += 2;
  }

  private static void output8 (int value) {
    if (pass2) {
      BigEndianMemory.write8 (dot.value, value);
      sourceeditor.objcode (sourcelinenum, dot.value, 1, value);
    }
    dot.value ++;
  }

  // Send error message back to source editor
  // Since undefined symbol errors may happen during pass1 and
  // get cleaned up during pass2, only output message in pass2

  private static void errormessage (String msg) {
    if (pass2) {

      // Put error message on source file window where object code goes

      haveerrors = true;
      sourceeditor.errormsg (sourcelinenum, msg);

      // Also display in a pop-up window (unless they've clicked Cancel)

      if (popuperrorsenabled) {

        // Start it with the error message itself

        StringBuffer where = new StringBuffer (msg);

        // If we're buried inside a macro expansion somewhere, tell the user where

        for (MacExp macexp = macroexpand; macexp != null; macexp = macexp.nextouter) {
          if (macexp.currentline != null) {
            where.append ("\n  macro ");
            where.append (macexp.mac.name);
            where.append (" ");
            where.append (macexp.lineindex);
            where.append (": ");
            where.append (macexp.currentline);
          }
        }

        // If we're inside include files somewhere, tell the user where

        for (SrcFile sf = srcfile; sf != null; sf = sf.outerfile) {
          if (sf.currentline != null) {
            where.append ("\n  include ");
            where.append (sf.filename);
            where.append (" ");
            where.append (sf.linenumber);
            where.append (": ");
            where.append (sf.currentline);
          }
        }

        // Anyway, tell the user which top-level source file the error is from

        where.append ("\n  ");
        where.append (sourceeditor.getsourcename ());
        where.append (" ");
        where.append (sourcelinenum);

        // Display pop-up error message.  If user clicks Cancel, don't display any more.

        if (JOptionPane.showConfirmDialog (null, where.toString (), "SparcAsm error", JOptionPane.OK_CANCEL_OPTION) == JOptionPane.CANCEL_OPTION) {
          popuperrorsenabled = false;
        }
      }
    }
  }
}
