//+++2004-11-01
//    Copyright (C) 2004  Mike Rieker, Beverly, MA USA
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//---2004-11-01

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class SparcCPU extends JInternalFrame {
  private int opcode;                   // opcode being executed

  private volatile boolean running;     // true when CPU is running
                                        // use 'setrunflag' routine to change it
  private Thread runthread;             // the sub-thread that is currently running

  private SparcCPU cpuframe;            // points to 'this'
  private JButton runbutton;            // various buttons
  private JButton haltbutton;
  private JButton stepbutton;
  private JButton resetbutton;

  public static final String EXC_BADMEMADDR    = "Bad memory address";
  public static final String EXC_UNALIGNEDADDR = "Unaligned address";
  public static final String EXC_ODDREGISTER   = "Odd register number";
  public static final String EXC_UNDEFOPCODE   = "Undefined opcode";
  public static final String EXC_NOMOREFRAMES  = "No more register frames";

  public final static int O6  = 14;       // register numbers
  public final static int O7  = 15;
  public final static int PC  = -1;
  public final static int NPC = -2;
  public final static int PS  = -3;
  public final static int Y   = -4;
  private final static int NN =  4;

  private int[] globalregisters;          // global register array [8]
  private int[] negativeregisters;        // negative register array [NN]
  private RegisterSet outermostregisters; // outermost register set
  private RegisterSet activeregisters;    // current register set for l0..l7;i0..i7
                                          // - for o0..o7, use .inner's i0..i7
  private RegisterSet displayedregisters; // displayed registers

  private int active_index;               // index of active register set
                                          // - 0=outermost

  private JScrollBar registerscrollbar;
  private HexTextField[] globalregistervalues;
  private HexTextField[] negativeregistervalues;
  private HexTextField[] windowregistervalues;

  // Constructor

  public SparcCPU () {
    super ("Sparc CPU", true, true, true, true);
    cpuframe = this;

    // Put the registers on top and the bottons on the bottom

    JPanel cpupanel = (JPanel) cpuframe.getContentPane ();
    cpupanel.setLayout (new BoxLayout (cpupanel, BoxLayout.Y_AXIS));

    cpupanel.add (constructregisterpanel ());
    cpupanel.add (constructbuttonpanel ());

    // Make it minimal size to hold everything and make it visible

    cpuframe.pack ();
    cpuframe.setVisible (true);

    // Put a little separation between the pieces

    Dimension cpuframesize = cpuframe.getSize ();  // get minimal size created by pack
    cpuframesize.height += 10;                     // add a few more pixels
    cpuframesize.width  += 10;
    cpuframe.setSize (cpuframesize);               // set the new size
    cpuframe.validate ();                          // ... then reformat
    cpuframe.doLayout ();

    // Reset CPU state

    reset ();
  }

  // Called by the constructor to construct 'registerpanel'
  // which holds the reigster display, including the scrollbar

  private JPanel constructregisterpanel () {
    JLabel[] globalregisterlabels;
    JLabel[] negativeregisterlabels;
    JLabel[] windowregisterlabels;

    // Initialize register arrays

    globalregisters = new int[8];                // global registers are just an array of 8 ints
    negativeregisters = new int[NN];             // negative registers are an array of NN ints
    activeregisters = new RegisterSet();         // set up outermost active native (%l0..%l7,%i0..%i7)
    activeregisters.inner = new RegisterSet ();  // borrowed registers for %o0..%o7
    activeregisters.inner.outer = activeregisters;
    displayedregisters = activeregisters;        // set displayed set = active set
    outermostregisters = activeregisters;
    active_index = 0;                            // outermost register set is active

    globalregisterlabels   = new JLabel[8];      // make array of labels for the boxes
    negativeregisterlabels = new JLabel[NN];
    windowregisterlabels   = new JLabel[24];
    globalregistervalues   = new HexTextField[8];  // these arrays have the actual text boxes
    negativeregistervalues = new HexTextField[NN];
    windowregistervalues   = new HexTextField[24];

    for (int i = 0; i < 8; i ++) {                             // fill in stuff that is 0..7
      globalregistervalues[i] = new HexTextField (8);          // - like the global register text fields
      globalregistervalues[i].addFocusListener (new RegisterFocusListener (0, i, globalregistervalues));
      globalregisterlabels[i]    = new JLabel ("  g" + i);     // register labels
      windowregisterlabels[i+ 0] = new JLabel ("  o" + i);
      windowregisterlabels[i+ 8] = new JLabel ("  l" + i);
      windowregisterlabels[i+16] = new JLabel ("  i" + i);
    }
    globalregistervalues[0].setEditable (false);               // don't allow editing %g0 (it's always zero)
    windowregisterlabels[6+ 0].setText ("  sp");               // %o6 is the stack pointer
    windowregisterlabels[6+16].setText ("  fp");               // %i6 is the frame pointer

    for (int i = 0; i < NN; i ++) {                            // fill in stuff for the negative registers
      negativeregistervalues[i] = new HexTextField (8);
      negativeregistervalues[i].addFocusListener (new RegisterFocusListener (-1, i, negativeregistervalues));
    }
    negativeregisterlabels[~PC]  = new JLabel ("pc");          // - each label is something special
    negativeregisterlabels[~NPC] = new JLabel ("npc");
    negativeregisterlabels[~PS]  = new JLabel ("ps");
    negativeregisterlabels[~Y]   = new JLabel ("y");

    for (int i = 0; i < 24; i ++) {                            // fill in stuff that is 0..23
      windowregistervalues[i] = new HexTextField (8);          // - such as the windowed register display boxes
      windowregistervalues[i].addFocusListener (new RegisterFocusListener (1, i, windowregistervalues));
    }

    // Set up register display panel

    JPanel registerpanel = new JPanel ();

    registerscrollbar = new JScrollBar (JScrollBar.HORIZONTAL, 0, 1, 0, 1);     // scrollbar used to select windowed register set
    registerscrollbar.addAdjustmentListener (new RegisterScrollbarListener ()); // set up a listener so we know when to switch registers

    GridBagLayout rlgbl = new GridBagLayout();                         // use grid bag layout to set up the panel
    GridBagConstraints rlgbc = new GridBagConstraints();
    registerpanel.setLayout (rlgbl);
    JLabel filler;

    rlgbc.fill = GridBagConstraints.BOTH;
    rlgbc.weightx = 1.0;

    for (int i = 0; (i < NN) || (i <= 8); i ++) {                      // loop through all rows
      if (i < NN) {                                                    // see if within the negative register set rows
        rlgbc.gridwidth = 1;                                           // ok, this stuff is one column wide each
        rlgbl.setConstraints (negativeregisterlabels[i], rlgbc);       //     output a label
        registerpanel.add (negativeregisterlabels[i]);
        rlgbl.setConstraints (negativeregistervalues[i], rlgbc);       //     output the contents
        registerpanel.add (negativeregistervalues[i]);
      } else {
        rlgbc.gridwidth = 1;                                           // beyond negative registers, ...
        filler = new JLabel (" ");                                     //     use filler for label spot
        rlgbl.setConstraints (filler, rlgbc);
        registerpanel.add (filler);
        filler = new JLabel (" ");                                     //     use filler for contents spot
        rlgbl.setConstraints (filler, rlgbc);
        registerpanel.add (filler);
      }
      if (i < 8) {                                                     // see if within general register set rows
        rlgbc.gridwidth = 1;                                           // ok, this stuff is one column wide each
        rlgbl.setConstraints (globalregisterlabels[i], rlgbc);         //     output a global register label
        registerpanel.add (globalregisterlabels[i]);
        rlgbl.setConstraints (globalregistervalues[i], rlgbc);         //     output the global register contents
        registerpanel.add (globalregistervalues[i]);
        for (int j = 0; j < 24; j += 8) {                              //     loop for o's, l's and i's
          if (j == 16) rlgbc.gridwidth = GridBagConstraints.RELATIVE;  //     i labels are next to last column
          rlgbl.setConstraints (windowregisterlabels[i+j], rlgbc);     //     output windowed register label
          registerpanel.add (windowregisterlabels[i+j]);
          if (j == 16) rlgbc.gridwidth = GridBagConstraints.REMAINDER; //     i contents are very last column
          rlgbl.setConstraints (windowregistervalues[i+j], rlgbc);     //     output windowed register contents
          registerpanel.add (windowregistervalues[i+j]);
        }
      } else {
        rlgbc.gridwidth = 1;                                           // beyond general registers, ...
        filler = new JLabel (" ");                                     //     use filler for global register labels
        rlgbl.setConstraints (filler, rlgbc);
        registerpanel.add (filler);
        filler = new JLabel (" ");                                     //     use filler for global register contents
        rlgbl.setConstraints (filler, rlgbc);
        registerpanel.add (filler);
        if (i == 8) {
          rlgbc.gridwidth = GridBagConstraints.REMAINDER;              //     on row 8 we have the scroll bar for the rest of row
          rlgbl.setConstraints (registerscrollbar, rlgbc);
          registerpanel.add (registerscrollbar);
        } else {
          for (int j = 0; j < 24; j += 8) {                            //     otherwise, use filler for windowed registers
            if (j == 16) rlgbc.gridwidth = GridBagConstraints.RELATIVE;
            filler = new JLabel (" ");
            rlgbl.setConstraints (filler, rlgbc);
            registerpanel.add (filler);
            if (j == 16) rlgbc.gridwidth = GridBagConstraints.REMAINDER;
            filler = new JLabel (" ");
            rlgbl.setConstraints (filler, rlgbc);
            registerpanel.add (filler);
          }
        }
      }
    }

    return (registerpanel);
  }

  // Construct the 'buttonpanel' containing the buttons

  private JPanel constructbuttonpanel () {

    // Arrange buttons horizontally in a little box to go along the bottom

    JPanel buttonpanel = new JPanel ();
    buttonpanel.setLayout (new BoxLayout (buttonpanel, BoxLayout.X_AXIS));

    runbutton = new JButton ("Run");
    runbutton.addActionListener (new RunButtonListener ());
    buttonpanel.add (runbutton);

    haltbutton = new JButton ("Halt");
    haltbutton.addActionListener (new HaltButtonListener ());
    buttonpanel.add (haltbutton);

    stepbutton = new JButton ("Step");
    stepbutton.addActionListener (new StepButtonListener ());
    buttonpanel.add (stepbutton);

    resetbutton = new JButton ("Reset");
    resetbutton.addActionListener (new ResetButtonListener ());
    buttonpanel.add (resetbutton);

    return (buttonpanel);
  }

  // Reset to initial state

  public void reset () {
    killoldthread ();           // if CPU is running in sub-thread, kill and wait for it

    // Reset registers

    activeregisters = outermostregisters;
    displayedregisters = outermostregisters;
    active_index = 0;
    registerscrollbar.setMaximum (1);
    for (int i = -NN; i < 32; i ++) setregister (i, 0);

    // Set current stack pointer [%o6] = end of memory

    setregister (O6, BigEndianMemory.getsize ());

    // Set current program counter [%pc] = symbol _start

    int startaddr = SparcAsm.getstartaddr ();
    setregister (PC, startaddr);
    setregister (NPC, startaddr + 4);

    // We aren't running, so enable Run and Step buttons, disable Halt button
    // This also updates the source editor windows to our current position

    setrunflag (false);
  }

  // Step one instruction

  // registerpc = address of instruction to execute this time
  // registernpc = address of instruction to execute next time

  public void step () throws CPUException {
    int pc  = getregister (PC);
    int npc = getregister (NPC);

    opcode = memread32 (pc);                       // read instruction from memory

majordecode:
    switch ((opcode >> 30) & 3) {                  // decode opcode<31:30>

      /* Branches */

      case 0: {
        boolean cond = false;

        switch ((opcode >> 22) & 0x7F) {	// decode opcode<28:22> (opcode<29> is annul bit)
          case 0x02: {				// bn
            cond = false;
            break;
          }
          case 0x0A: {				// be
            cond = testzbit ();
            break;
          }
          case 0x12: {                         // ble
            cond = (testnbit () ^ testvbit ()) || testzbit ();
            break;
          }
          case 0x1A: {                         // bl
            cond = testnbit () ^ testvbit ();
            break;
          }
          case 0x22: {                         // bleu
            cond = testcbit () || testzbit ();
            break;
          }
          case 0x2A: {                         // blu,bcs
            cond = testcbit ();
            break;
          }
          case 0x32: {                         // bneg
            cond = testnbit ();
            break;
          }
          case 0x3A: {                         // bvs
            cond = testvbit ();
            break;
          }
          case 0x42: {                         // ba
            cond = true;
            break;
          }
          case 0x4A: {                         // bne
            cond = !testzbit ();
            break;
          }
          case 0x52: {                         // bg
            cond = !((testnbit () ^ testvbit ()) || testzbit ());
            break;
          }
          case 0x5A: {                         // bge
            cond = !(testnbit () ^ testvbit ());
            break;
          }
          case 0x62: {                         // bgu
            cond = !(testcbit () || testzbit ());
            break;
          }
          case 0x6A: {                         // bgeu,bcc
            cond = !testcbit ();
            break;
          }
          case 0x72: {                         // bpos
            cond = !testnbit ();
            break;
          }
          case 0x7A: {                         // bvc
            cond = !testvbit ();
            break;
          }
          case 0x04: case 0x0C: case 0x14: case 0x1C:
          case 0x24: case 0x2C: case 0x34: case 0x3C:
          case 0x44: case 0x4C: case 0x54: case 0x5C:
          case 0x64: case 0x6C: case 0x74: case 0x7C:
          {                                    // sethi
            setregister ((opcode >> 25) & 0x1F, opcode << 10);
            pc = npc;
            npc += 4;
            break majordecode;
          }
          default: {
            cpuexception (EXC_UNDEFOPCODE);
          }
        }
        if (cond) {                            // if branch condition, ...
          int brdest = pc + ((opcode << 10) >> 8);
          pc = npc;                            // execute delay slot instr next
          npc = brdest;                        // followed by target instruction
          if ((opcode & 0x3FC00000) == 0x30800000) { // check for ba,a
            pc = brdest;                       // ok, don't execute delay slot instr
            npc = pc + 4;
          }
        } else {
          pc = npc;
          if ((opcode & 0x20000000) != 0) pc += 4; // maybe skip annulled instr
          npc = pc + 4;                        // then next instr directly follows
        }
        break;
      }

      /* Call */

      case 1: {
        int destination;

        destination = pc + (opcode << 2);
        setregister (O7, pc);	// save call address in %o7
        pc = npc;
        npc = destination;
        break;
      }

      /* Arithmetic */

      case 2: {
        int op1 = getregister ((opcode >> 14) & 0x1F);
        int op2 = getarithop2 (opcode);
        int rd  = (opcode >> 25) & 0x1F;

        switch ((opcode >> 19) & 0x3F) {

          // ADD

          case 0x00: {
            int out = op1 + op2;
            setregister (rd, out);
            break;
          }

          // AND

          case 0x01: {
            int out = op1 & op2;
            setregister (rd, out);
            break;
          }

          // OR

          case 0x02: {
            int out = op1 | op2;
            setregister (rd, out);
            break;
          }

          // XOR

          case 0x03: {
            int out = op1 ^ op2;
            setregister (rd, out);
            break;
          }

          // SUB

          case 0x04: {
            int out = op1 - op2;
            setregister (rd, out);
            break;
          }

          // ANDN

          case 0x05: {
            int out = op1 & ~ op2;
            setregister (rd, out);
            break;
          }

          // ORN

          case 0x06: {
            int out = op1 | ~ op2;
            setregister (rd, out);
            break;
          }

          // XNOR

          case 0x07: {
            int out = op1 ^ ~ op2;
            setregister (rd, out);
            break;
          }

          // ADDX

          case 0x08: {
            int out = op1 + op2 + getcbit ();
            setregister (rd, out);
            break;
          }

          // SUBX

          case 0x0C: {
            int out = op1 - op2 - getcbit ();
            setregister (rd, out);
            break;
          }

          // ADDCC

          case 0x10: {
            int out = op1 + op2;
            setregister (rd, out);
            setnzvc (out < 0,
                     out == 0,
                     (op1 < 0 && op2 < 0 && out >= 0) ||
                            (op1 >= 0 && op2 >= 0 && out < 0),
                     (op1 < 0 && op2 < 0) ||
                            (out >= 0 && (op1 < 0 || op2 < 0)));
            break;
          }

          // ANDCC

          case 0x11: {
            int out = op1 & op2;
            setregister (rd, out);
            setnzvc (out < 0, out == 0, false, false);
            break;
          }

          // ORCC

          case 0x12: {
            int out = op1 | op2;
            setregister (rd, out);
            setnzvc (out < 0, out == 0, false, false);
            break;
          }

          // XORCC

          case 0x13: {
            int out = op1 ^ op2;
            setregister (rd, out);
            setnzvc (out < 0, out == 0, false, false);
            break;
          }

          // SUBCC

          case 0x14: {
            int out = op1 - op2;
            setregister (rd, out);
            setnzvc (out < 0,
                     out == 0,
                     (op1 < 0 && op2 >= 0 && out >= 0) ||
                            (op1 >= 0 && op2 < 0 && out < 0),
                     (op1 >= 0 && op2 < 0) ||
                            (out < 0 && (op1 >= 0 || op2 < 0)));
            break;
          }

          // ANDNCC

          case 0x15: {
            int out = op1 & ~ op2;
            setregister (rd, out);
            setnzvc (out < 0, out == 0, false, false);
            break;
          }

          // ORNCC

          case 0x16: {
            int out = op1 | ~ op2;
            setregister (rd, out);
            setnzvc (out < 0, out == 0, false, false);
            break;
          }

          // XNORCC

          case 0x17: {
            int out = op1 ^ ~ op2;
            setregister (rd, out);
            setnzvc (out < 0, out == 0, false, false);
            break;
          }

          // ADDXCC

          case 0x18: {
            int out = op1 + op2 + getcbit ();
            setregister (rd, out);
            setnzvc (out < 0,
                     out == 0,
                     (op1 < 0 && op2 < 0 && out >= 0) ||
                            (op1 >= 0 && op2 >= 0 && out < 0),
                     (op1 < 0 && op2 < 0) ||
                            (out >= 0 && (op1 < 0 || op2 < 0)));
            break;
          }

          // SUBXCC

          case 0x1C: {
            int out = op1 - op2 - getcbit ();
            setregister (rd, out);
            setnzvc (out < 0,
                     out == 0,
                     (op1 < 0 && op2 >= 0 && out >= 0) ||
                            (op1 >= 0 && op2 < 0 && out < 0),
                     (op1 >= 0 && op2 < 0) ||
                            (out < 0 && (op1 >= 0 || op2 < 0)));
            break;
          }

          // MULSCC

          case 0x24: {
            int out = ((op1 >> 1) & 0x7FFFFFFF) | ((getnbit () ^ getvbit ()) << 31);
            int y = getregister (Y);

            if ((y & 1) != 0) out += op2;
            y = ((y >> 1) & 0x7FFFFFFF) | (op1 << 31);

            setregister (rd, out);
            setregister (Y, y);
            setnzvc (out < 0,
                     out == 0,
                     (op1 < 0 && op2 < 0 && out >= 0) ||
                            (op1 >= 0 && op2 >= 0 && out < 0),
                     (op1 < 0 && op2 < 0) ||
                            (out >= 0 && (op1 < 0 || op2 < 0)));
            break;
          }

          // SLL

          case 0x25: {
            int out = op1 << (op2 & 31);
            setregister (rd, out);
            break;
          }

          // SRL

          case 0x26: {
            int out = op1 >> (op2 & 31);
            out &= (1 << (32 - (op2 & 31))) - 1;
            setregister (rd, out);
            break;
          }

          // SRA

          case 0x27: {
            int out = op1 >> (op2 & 31);
            setregister (rd, out);
            break;
          }

          // RDY

          case 0x28: {
            setregister (rd, getregister (Y));
            break;
          }

          // WRY

          case 0x30: {
            setregister (Y, op1 ^ op2);
            break;
          }

          // JMPL

          case 0x38: {
            setregister (rd, pc);
            pc = npc;
            npc = op1 + op2;
            break majordecode;
          }

          // SAVE

          case 0x3C: {
            int out = op1 + op2;
            saveregisters ();
            setregister (rd, out);
            break;
          }

          // RESTORE

          case 0x3D: {
            int out = op1 + op2;
            if (!restoreregisters ()) cpuexception (EXC_NOMOREFRAMES);
            setregister (rd, out);
            break;
          }

          // Undefined/Unsupported

          default: {
            cpuexception (EXC_UNDEFOPCODE);
          }
        }

        /* Increment program counter linearly as all arith instructions simply increment to the next instruction */

        pc = npc;
        npc += 4;
        break;
      }

      /* Loads and Stores */

      case 3: {
        switch ((opcode >> 19) & 0x3F) {
          case 0x01: {		// ldub
            setregister ((opcode >> 25) & 0x1F, memread8 (getmemaddr (opcode)) & 0xFF);
            break;
          }
          case 0x09: {		// ldsb
            setregister ((opcode >> 25) & 0x1F, memread8 (getmemaddr (opcode)));
            break;
          }
          case 0x02: {		// lduh
            setregister ((opcode >> 25) & 0x1F, memread16 (getmemaddr (opcode)) & 0xFFFF);
            break;
          }
          case 0x0A: {		// ldsh
            setregister ((opcode >> 25) & 0x1F, memread16 (getmemaddr (opcode)));
            break;
          }
          case 0x00: {		// ld
            setregister ((opcode >> 25) & 0x1F, memread32 (getmemaddr (opcode)));
            break;
          }
          case 0x03: {		// ldd
            int memoryaddress = getmemaddr (opcode);
            int registernumber = (opcode >> 25) & 0x1F;
            int data0, data1;

            if ((registernumber & 1) != 0) cpuexception (EXC_ODDREGISTER);
            if ((memoryaddress  & 4) != 0) cpuexception (EXC_UNALIGNEDADDR);
            data0 = memread32 (memoryaddress + 0);
            data1 = memread32 (memoryaddress + 4);
            setregister (registernumber + 0, data0);
            setregister (registernumber + 4, data1);
            break;
          }
          case 0x05: {		// stb
            memwrite8 (getmemaddr (opcode), getregister ((opcode >> 25) & 0x1F));
            break;
          }
          case 0x06: {		// sth
            memwrite16 (getmemaddr (opcode), getregister ((opcode >> 25) & 0x1F));
            break;
          }
          case 0x04: {		// st
            memwrite32 (getmemaddr (opcode), getregister ((opcode >> 25) & 0x1F));
            break;
          }
          case 0x07: {		// std
            int memoryaddress = getmemaddr (opcode);
            int registernumber = (opcode >> 25) & 0x1F;

            if ((registernumber & 1) != 0) cpuexception (EXC_ODDREGISTER);
            if ((memoryaddress  & 4) != 0) cpuexception (EXC_UNALIGNEDADDR);
            memwrite32 (memoryaddress + 0, getregister (registernumber + 0));
            memwrite32 (memoryaddress + 4, getregister (registernumber + 1));
            break;
          }
          case 0x0F: {		// swap
            int memoryaddress = getmemaddr (opcode);
            int registernumber = (opcode >> 25) & 0x1F;
            int temp;

            temp = memread32 (memoryaddress);                          // read memory into temp
            memwrite32 (memoryaddress, getregister (registernumber));  // write register out to memory
            setregister (registernumber, temp);                        // write register last in case memwrite32 faults
            break;
          }
          default: {
            cpuexception (EXC_UNDEFOPCODE);
          }
        }

        /* Increment program counter linearly */

        pc = npc;
        npc += 4;
        break;
     }

      /* Unknown opcode */

      default: {
        cpuexception (EXC_UNDEFOPCODE);
        break;
      }
    }

    // Update 'permanent' PC and NPC

    setregister (PC,  pc);
    setregister (NPC, npc);
  }

  /* Extract the value of 'op2' for an arithmetic instruction                     */
  /* The value may be iether from a register or an immediate in the opcode itself */

  private int getarithop2 (int opcode) {
    if ((opcode & 0x2000) == 0) return (getregister (opcode & 0x1F));	// if i is clear, return the indicated register number
    opcode <<= 32 - 13;				// i is set, shift sign bit to top opcode bit
    return (opcode >> 19);			// shift value back down in place, sign extending it
  }

  /* Get memory address given in an load/store opcode */

  private int getmemaddr (int opcode) {

    // If opcode<13> is set, it's a register + signed immediate value

    if ((opcode & 0x2000) != 0) {
      return (getregister ((opcode >> 14) & 0x1F)
            + ((opcode << 19) >> 19));
    }

    // Otherwise, it's the sum of two registers

    else {
      return (getregister ((opcode >> 14) & 0x1F)
            + getregister (opcode & 0x1F));
    }
  }

  /* Read & Write condition codes in the PS */

  private int getnbit () {
    return ((getregister (PS) >> 23) & 1);
  }
  private boolean testnbit () {
    return ((getregister (PS) & 0x800000) != 0);
  }

  private boolean testzbit () {
    return ((getregister (PS) & 0x400000) != 0);
  }

  private int getvbit () {
    return ((getregister (PS) >> 21) & 1);
  }
  private boolean testvbit () {
    return ((getregister (PS) & 0x200000) != 0);
  }

  private int getcbit () {
    return ((getregister (PS) >> 20) & 1);
  }
  private boolean testcbit () {
    return ((getregister (PS) & 0x100000) != 0);
  }

  private void setnzvc (boolean nbit, boolean zbit, boolean vbit, boolean cbit) {
    int ps = getregister (PS) & ~0xF00000;
    if (nbit) ps |= 0x800000;
    if (zbit) ps |= 0x400000;
    if (vbit) ps |= 0x200000;
    if (cbit) ps |= 0x100000;
    setregister (PS, ps);
  }

  /* Read & Write memory, checking for proper alignment                       */
  /* They should also, at some point, perform virtual-to-physical translation */

  private int memread32 (int address) throws CPUException {
    if ((address & 3) != 0) cpuexception (EXC_UNALIGNEDADDR);
    return (BigEndianMemory.read32 (address));
  }

  private int memread16 (int address) throws CPUException {
    if ((address & 1) != 0) cpuexception (EXC_UNALIGNEDADDR);
    return (BigEndianMemory.read16 (address));
  }

  private int memread8 (int address) throws CPUException {
    return (BigEndianMemory.read8 (address));
  }

  private void memwrite32 (int address, int value) throws CPUException {
    if ((address & 3) != 0) cpuexception (EXC_UNALIGNEDADDR);
    BigEndianMemory.write32 (address, value);
  }

  private void memwrite16 (int address, int value) throws CPUException {
    if ((address & 1) != 0) cpuexception (EXC_UNALIGNEDADDR);
    BigEndianMemory.write16 (address, value);
  }

  private void memwrite8 (int address, int value) throws CPUException {
    BigEndianMemory.write8 (address, value);
  }

  // Throw an exception for the given reason

  private void cpuexception (String reason) throws CPUException {
    throw new CPUException (getregister (PC), opcode, reason);
  }

  // Set the 'running' flag state
  // This must be threadsafe

  private void setrunflag (boolean rf) {

    // If we are now running, disable the Run and Step buttons, enable the Halt and Reset buttons

    if (rf) {
      runbutton.setEnabled (false);
      haltbutton.setEnabled (true);
      stepbutton.setEnabled (false);
      resetbutton.setEnabled (true);
    }

    // If we are now stopped, enable the Run, Step and Reset buttons, disable the Halt button

    else {
      runbutton.setEnabled (true);
      haltbutton.setEnabled (false);
      stepbutton.setEnabled (true);
      resetbutton.setEnabled (true);
      SparcSim.updateselocation ();
    }

    // Save flag's new value

    running = rf;
  }

  // Reset button listener - resets the CPU to original state

  private class ResetButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      reset ();
    }
  }

  // Step button listener - steps the CPU one instruction

  private class StepButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      killoldthread ();                  // there shouldn't be any old thread as the Step button
                                         // ... should be disabled when the thread is running
      setrunflag (true);                 // disable the Run and Step buttons whilst stepping
      try {
        step ();                         // step a single instruction
      } catch (CPUException sce) {
        JOptionPane.showMessageDialog (cpuframe, "SparcCPU: Fatal exception\n"
                                       + sce.reason + "\n  At PC="
                                       + HexTextField.hexstring (sce.registerpc)
                                       + "\n  Opcode=" + HexTextField.hexstring (sce.opcode));
      }
      setrunflag (false);                 // enable the Run and Step buttons
    }
  }

  // Run button listener - steps the CPU until stopped by an exception, breakpoint or the Halt button

  private class RunButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {

      // If there is a sub-thread already going, tell it to stop and wait for it to terminate

      killoldthread ();

      // Start a new thread going.  Disable all but the Halt and Reset buttons.

      RunThread rt = new RunThread ();  // create thread object
      setrunflag (true);                // disable the Run and Step buttons
      runthread = rt;                   // remember what thread to wait for
      rt.start ();                      // start the thread going
    }
  }

  // This is the thread that runs as a result of the Run button
  // It will keep going until it hits a breakpoint, there is an exception
  // ... or it sees that the Halt button has reset the 'running' flag

  private class RunThread extends Thread {
    public void run () {
      try {
        do step();
        while (running &&
               !SparcSim.checkbreakpoints (getregister (PC)));
      } catch (CPUException sce) {
        JOptionPane.showMessageDialog (cpuframe, "SparcCPU: Fatal exception\n"
                                       + sce.reason + "\n  At PC="
                                       + HexTextField.hexstring (sce.registerpc)
                                       + "\n  Opcode=" + HexTextField.hexstring (sce.opcode));
      }
      setrunflag (false);
    }
  }

  // This is called when the Halt button is clicked
  // It simply tells the sub-thread to stop stepping then waits for it to die

  private class HaltButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      killoldthread ();
    }
  }

  // If there is a sub-thread, this routine tells it to stop then waits for it to die

  private void killoldthread () {
    if (runthread != null) {
      running = false;
      while (runthread.isAlive ()) {
        try {
          runthread.join ();
        } catch (InterruptedException ie) {
          JOptionPane.showMessageDialog (cpuframe, "Waiting for old thread but it was interrupted");
        }
      }
      runthread = null;
    }
  }

  // Read and Write the active register set; update display if needed

  public int getregister (int regnum) {
    if (regnum < 0) return (negativeregisters[~regnum]);           // pc,npc,ps,y
    if (regnum < 8) return (globalregisters[regnum]);              // r0..r7 = g0..g7
    if (regnum < 16) return (activeregisters.inner.regs[regnum]);  // r8..r15 = o0..o7 = borrowed i0..i7
    return (activeregisters.regs[regnum-16]);                      // r16..r31 = l0..l7,i0..i7 = native l0..l7,i0..i7
  }

  public void setregister (int regnum, int value) {
    if (regnum == 0) return;                                       // don't allow changing %g0
    if (regnum < 0) {
      negativeregisters[~regnum] = value;                          // negative registers, update binary value
      negativeregistervalues[~regnum].setBinary (value);           //                     update display
    } else if (regnum < 8) {
      globalregisters[regnum] = value;                             // global registers, update binary value
      globalregistervalues[regnum].setBinary (value);              //                   update display
    } else if (regnum < 16) {
      activeregisters.inner.regs[regnum] = value;                  // output register, update inner input register
      if (activeregisters.inner == displayedregisters.inner) {     //                  maybe update display
        windowregistervalues[regnum-8].setBinary (value);
      }
      if (activeregisters.inner == displayedregisters) {
        windowregistervalues[regnum-8].setBinary (value);
      }
    } else {
      activeregisters.regs[regnum-16] = value;                     // local or input register, update register directly
      if (activeregisters == displayedregisters) {                 //                          maybe update display
        windowregistervalues[regnum-8].setBinary (value);
      }
      if ((activeregisters == displayedregisters.inner) && (regnum >= 24)) {
        windowregistervalues[regnum-24].setBinary (value);
      }
    }
  }

  // Push active registers, ie, nest in a level

  private void saveregisters () {
    activeregisters = activeregisters.inner;            // link to inner frame
    if (activeregisters.inner == null) {                // see if it has an inner frame
      activeregisters.inner = new RegisterSet ();       // (so there will always be room ...
      activeregisters.inner.outer = activeregisters;    //  ... for %o0..%o7)
    }
    active_index ++;                                    // we're in a deeper register set
    registerscrollbar.setMaximum (active_index + 1);    // scrollbar goes from 0..active_index
  }

  // Pop active registers, ie, nest out a level

  private boolean restoreregisters () {
    if (activeregisters.outer == null) return (false);  // can't if already at outermost level
    activeregisters = activeregisters.outer;
    active_index --;                                    // we're in an outer register level
    registerscrollbar.setMaximum (active_index + 1);    // scrollbar goes from 0..active_index
    return (true);
  }

  // An object of this class is created to process register scrollbar changes
  // When called, we change which windowed register set is being displayed

  // Scrollbar returns zero on the left end and active_index on the right end

  private class RegisterScrollbarListener implements AdjustmentListener {
    public void adjustmentValueChanged (AdjustmentEvent e) {

      // Calculate displayed_index
      // It's zero for active register set
      // It's maximum-1 for outermost register set

      int displayed_index = active_index - registerscrollbar.getValue ();

      // Point displayedregisters at the correct register set

      displayedregisters = outermostregisters;
      for (int i = displayed_index; -- i >= 0;) {
        displayedregisters = displayedregisters.inner;
      }

      // Display whatever displayedregisters ends up pointing to

      for (int i = 0; i < 8; i ++) {
        windowregistervalues[i+ 0].setBinary (displayedregisters.inner.regs[i+8]); // %o0..%o7=next inner %i0..%i7
        windowregistervalues[i+ 8].setBinary (displayedregisters.regs[i+0]);       // %l0..%l7
        windowregistervalues[i+16].setBinary (displayedregisters.regs[i+8]);       // %i0..%i7
      }
    }
  }

  // These objects are created to monitor when the user has manually modified a register
  // There is one per register content text box

  // Store new value in register binary value array
  // Also, if they change the PC, set the NPC to the new PC+4

  private class RegisterFocusListener extends FocusAdapter {
    private int domain;
    private int index;
    private HexTextField[] textfields;

    // d = -1: negative register; 0<=i<NN
    //      0: global register;   1<=i< 8
    //      1: windowed register; 0<=i< 8 : native %o0..%o7 = borrowed %i0..%i7
    //                            8<=i<16 : native %l0..%l7
    //                           16<=i<24 : native %i0..%i7
    // tf = text fields array (indexed by i)

    public RegisterFocusListener (int d, int i, HexTextField[] tf) {
      domain = d;
      index = i;
      textfields = tf;
    }

    // This gets called when clicking out of the text box
    // (presumably after having made changes)

    public void focusLost (FocusEvent fe) {

      // Store in register binary value array

      int value = textfields[index].getBinary ();
      switch (domain) {
        case -1: {
          negativeregisters[index] = value;
          if (index == ~PC) {
            negativeregisters[~NPC] = value + 4;
            negativeregistervalues[~NPC].setBinary (value + 4);
          }
          break;
        }
        case 0: {
          globalregisters[index] = value;
          break;
        }
        case 1: {
          if (index < 8) {
            displayedregisters.inner.regs[index+8] = value;
          } else {
            displayedregisters.regs[index-8] = value;
          }
          break;
        }
      }
    }
  }

  // Array of 16 windowed registers
  // [0..7] = %l0..%l7; [8..15] = %i0..%i7
  // To get %o0..%o7, you must 'borrow' the .inner %i0..%i7

  private class RegisterSet {
    public RegisterSet inner;   // link to inner register set
    public RegisterSet outer;   // link to outer register set
    public int[] regs;          // array of 16 values
    public RegisterSet () {
      regs = new int[16];
    }
  }

  /* This class is thrown by things when an CPU exception is      */
  /* encountered, such as an illegal opcode or bad memory address */

  public class CPUException extends Exception {
    public int registerpc;
    public int opcode;
    public String reason;

    // registerpc = program counter of opcode that got the exception
    // opcode     = opcode that got the exception
    // reason     = descriptive reason for the exception

    public CPUException (int registerpc, int opcode, String reason) {
      this.registerpc = registerpc;
      this.opcode = opcode;
      this.reason = reason;
    }
  }
}
